#!/usr/bin/env python

"""
Functions relating to the 'data' endpoint
"""

# Standard imports
from pathlib import Path
from typing import Optional, Union

# Libica api imports
from libica.openapi.v3 import ApiClient, ApiException
from libica.openapi.v3.api.data_api import DataApi
from urllib.parse import urlunparse, urlparse

# Libica model imports
from libica.openapi.v3.models import (
    Data,
    ProjectData
)
from pydantic import UUID4

# Local imports
from ...project_data import is_data_id_format
from ...utils.configuration import get_icav2_configuration
from ...utils.logger import get_logger
from ...utils.globals import (
    FOLDER_DATA_TYPE,
    FILE_DATA_TYPE,
    ICAV2_URI_SCHEME
)

# Log imports
logger = get_logger()


def get_data_obj_from_data_id(
        data_id: Union[UUID4, str],
        region_id: Optional[Union[UUID4, str]] = None
) -> Data:
    """
    Get data object by id

    :param data_id:
    :param region_id:

    :return: The data object
    :rtype: `Data <https://umccr.github.io/libica/openapi/v3/docs/Data/>`_

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.data import get_data_obj_by_id

        # Set vars
        data_id = "fil.123456"

        # Get data object by id
        data_obj = get_data_obj_by_id(data_id)
    """
    from ...region.functions.region_functions import get_default_region
    if region_id is None:
        region_id = get_default_region().id

    # Get the data urn
    data_urn = f"urn:ilmn:ica:region:{region_id}:data:{data_id}"

    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = DataApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve a data.
        api_response: Data = api_instance.get_data(data_urn)
    except ApiException as e:
        logger.error("Exception when calling DataApi->get_data: %s\n" % e)
        raise ApiException

    return api_response


def get_owning_project_id(
        data_id: Union[UUID4, str],
        region_id: Optional[Union[UUID4, str]] = None
) -> str:
    """
    Get the owning project id of a data object

    :param data_id:
    :param region_id:

    :return: The owning project id
    :rtype: str

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.data import get_owning_project_id

        # Set vars
        data_id = "fil.123456"

        # Get owning project id
        owning_project_id = get_owning_project_id(data_id)

    """
    data_obj = get_data_obj_from_data_id(data_id, region_id)
    return str(data_obj.details.owning_project_id)


def get_project_data_obj_from_data_id(
        data_id: Union[UUID4, str]
) -> ProjectData:
    """
    Get the project data object from a data id

    :param data_id:  The data id

    :return: The project data object
    :rtype: `ProjectData <https://umccr.github.io/libica/openapi/v3/docs/ProjectData/>`_

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.data import get_project_data_obj_from_data_id

        # Set vars
        data_id = "fil.123456"

        # Get project data object from just data id
        project_data_obj = get_project_data_obj_from_data_id(data_id)
    """
    # Local imports to prevent circular dependency
    from ...project_data import get_project_data_obj_by_id

    # Get the owning project id
    project_id = get_owning_project_id(data_id)

    return get_project_data_obj_by_id(project_id, data_id)


def convert_icav2_uri_to_data_obj(
        data_uri: str,
        create_data_if_not_found: bool = False
) -> Data:
    """
    Given a data uri, convert to a data object

    :param data_uri:  The data uri
    :param create_data_if_not_found:  If true, create the data if not found

    :return: The data object
    :rtype: `Data <https://umccr.github.io/libica/openapi/v3/docs/Data/>`_

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.data import convert_icav2_uri_to_data_obj

        # Set vars
        data_uri = "icav2://project_id/path/to/data_obj"

        # Convert to data object
        data_obj = convert_icav2_uri_to_data_obj(data_uri)
    """
    # Local imports to prevent circular dependency
    from ...project_data import convert_icav2_uri_to_project_data_obj

    # Convert to project data object
    project_data_obj = convert_icav2_uri_to_project_data_obj(
        data_uri=data_uri,
        create_data_if_not_found=create_data_if_not_found
    )

    # Return as data object
    return project_data_obj.data


def convert_data_obj_to_icav2_uri(data_obj: Data) -> str:
    """
    Given a data object, convert to a data uri

    :param data_obj:
    :return: The data uri
    :rtype: str

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.data import convert_data_obj_to_icav2_uri

        # Set vars
        data_obj = get_data_obj_from_data_id("fil.123456")

        # Convert to data uri
        data_uri = convert_data_obj_to_icav2_uri(data_obj)

    """
    return str(
        urlunparse(
            (
                ICAV2_URI_SCHEME,
                str(data_obj.details.owning_project_id),
                data_obj.details.path,
                None, None, None
            )
        )
    )


def coerce_data_id_path_or_icav2_uri_to_data_obj(
        data_id_path_or_uri: str,
        create_data_if_not_found: bool = False
) -> Optional[Data]:
    """
    Given a data id or uri, convert to a data object

    :param data_id_path_or_uri:
    :param create_data_if_not_found:  If true, create the data if not found (only applicable to paths or uris)

    :return: The data object
    :rtype: `Data <https://umccr.github.io/libica/openapi/v3/docs/Data/>`_

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.data import coerce_data_id_or_icav2_uri_to_data_obj

        # Set vars
        data_id_path_or_uri = "fil.123456"  # Or icav2://project-id/path/to/file  # Or  /path/to/file

        # Coerce to data object
        data_obj = coerce_data_id_or_icav2_uri_to_data_obj(data_id_path_or_uri)
    """
    from ...project_data import get_project_data_obj_from_project_id_and_path
    from ...project import get_project_id

    if is_data_id_format(data_id_path_or_uri):
        # Data ID, easy to convert across
        return get_data_obj_from_data_id(
            data_id=data_id_path_or_uri
        )
    elif urlparse(data_id_path_or_uri).scheme == ICAV2_URI_SCHEME:
        # ICAv2 URI, convert to data object
        return convert_icav2_uri_to_data_obj(
            data_uri=data_id_path_or_uri,
            create_data_if_not_found=create_data_if_not_found
        )
    else:
        # Data Path, convert to data object
        # Not as straight forward, need to first find this data, then convert to data object
        project_id = get_project_id()
        if Path(data_id_path_or_uri) == Path("/"):
            # There is no data id for the root directory (nor can we create one), we return none
            return None

        project_data_obj = get_project_data_obj_from_project_id_and_path(
            project_id=project_id,
            data_path=Path(data_id_path_or_uri),
            data_type=FOLDER_DATA_TYPE if data_id_path_or_uri.endswith("/") else FILE_DATA_TYPE,
            create_data_if_not_found=create_data_if_not_found
        )
        return project_data_obj.data



#!/usr/bin/env python3

"""
List of available functions:

"""
# Standard imports
import re
import warnings
from io import TextIOWrapper
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict, List, Union, Optional, Any, Tuple, cast
from datetime import datetime
from urllib.parse import urlparse, urlunparse
import requests

# Libica Api imports
from libica.openapi.v2.model.create_project_data_move_batch import CreateProjectDataMoveBatch
from libica.openapi.v3 import (
    ApiException,
    CreateFileAndUploadUrl,
    ProjectFileAndUploadUrl,
    ProjectDataMoveBatchApi,
    CreateProjectDataMoveBatchItem
)
from libica.openapi.v3.api.project_data_api import ProjectDataApi
from libica.openapi.v3.api.project_data_copy_batch_api import ProjectDataCopyBatchApi

# Libica model imports
from libica.openapi.v3.models import (
    AnalysisInputExternalData,
    AwsTempCredentials,
    CreateData,
    CreateProjectDataCopyBatch,
    CreateProjectDataCopyBatchItem,
    CreateTemporaryCredentials,
    DataIdOrPathList,
    DataUrlWithPath,
    Download,
    Job,
    ProjectData,
    ProjectDataCopyBatch,
    TempCredentials,
    Upload
)
from libica.openapi.v3 import ApiClient
from libica.openapi.v3 import (
    CreateFileData, CreateFolder,
)
from pydantic import UUID4

# Local imports
from ...literals import (
    DataType,
    ProjectDataSortParameterType,
    ProjectDataStatusValuesType,
    UriType,
    DataTagType
)
from ...utils.configuration import get_icav2_configuration, logger
from ...utils.globals import (
    LIBICAV2_DEFAULT_PAGE_SIZE,
    IS_REGEX_MATCH,
    FILE_DATA_TYPE,
    FOLDER_DATA_TYPE,
    ICAV2_URI_SCHEME,
    S3_URI_SCHEME
)
from ...utils.miscell import is_uuid_format, is_uri_format


def get_project_data_file_id_from_project_id_and_path(
        project_id: Union[UUID4, str],
        file_path: Path,
        create_file_if_not_found: bool = False
) -> str:
    """
    Given a project id, parent folder path and file_name, return the file id
    If the file is not found, and create_file_if_not_found is True, create the file

    :param project_id:  The project id
    :param file_path:  The path to the file
    :param create_file_if_not_found:  Create the file object if it does not exist

    :return: The file id

    :raises: FileNotFoundError, ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import (
            create_download_urls, get_project_folder_id_from_project_id_and_path
        )
        from wrapica.libica_models import DataUrlWithPath

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        file_id: str = get_project_file_id_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            file_path=Path("/path/to/file.txt")
        )

        download_urls: List[DataUrlWithPath] = create_download_urls(
            project_id="proj.abcdef1234567890",
            folder_id=project_folder_obj.data.id,
            recursive=True
        )

        for download_url in download_urls:
            print(download_url.url)

    """
    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    parent_folder_path = str(file_path.parent.absolute()) + "/"
    if parent_folder_path == "//":
        parent_folder_path = "/"

    # Add the filename to the list of filenames to search on
    filename = [
        file_path.name
    ]

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the list of project data.
        data_items: List[ProjectData] = api_instance.get_project_data_list(
            project_id=str(project_id),
            parent_folder_path=parent_folder_path,
            filename=filename,
            filename_match_mode="EXACT",
            file_path_match_mode="FULL_CASE_INSENSITIVE",
            type=FILE_DATA_TYPE
        ).items
    except ApiException as e:
        if not create_file_if_not_found:
            logger.error("Exception when calling ProjectDataApi->get_project_data_list: %s\n" % e)
            raise ApiException
        else:
            file_obj = create_file_in_project(
                project_id=project_id,
                file_path=file_path.absolute(),
            )
            return file_obj.data.id

    # Get the file id
    try:
        file_id = next(
            filter(
                lambda data_iter: data_iter.data.details.path == str(file_path),
                data_items
            )
        )
    except StopIteration as e:
        if create_file_if_not_found:
            # Create the folder
            file_id = create_file_in_project(
                project_id=project_id,
                file_path=file_path
            )
        else:
            logger.error("Could not find file id for file: %s\n" % file_path)
            raise FileNotFoundError(f"Could not find file id for file: {file_path}") from e

    return file_id.data.id


def create_data_in_project(
        project_id: Union[UUID4, str],
        parent_folder_path: Path,
        data_name: str,
        data_type: DataType
) -> ProjectData:
    """
    Create a data object in a project context

    :param project_id: The project ID
    :param parent_folder_path: The parent folder path of where the data object needs to be created
    :param data_name:  The name of the file or folder
    :param data_type:  One of "FILE" or "FOLDER"

    :return: The newly created project data object
    :rtype: List[`ProjectData <https://umccr-illumina.github.io/libica/openapi/v3/docs/ProjectData/>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import (
            create_data_in_project, create_data_in_project,
        )
        from wrapica.enums import DataType
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_obj: ProjectData = create_data_in_project(
            project_id="abcd-1234-efab-5678",
            parent_folder_path=Path("/path/to/folder/"),
            data_name="file.txt",
            data_type="FILE"
        )
    """
    warnings.warn(
        "create_data_in_project is deprecated and will be removed in a future release. Use create_file_in_project or create_folder_in_project instead.",
        DeprecationWarning,
        stacklevel=2
    )

    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    parent_folder_path = str(parent_folder_path.absolute()) + "/"
    if parent_folder_path == "//":
        parent_folder_path = "/"

    # example passing only required values which don't have defaults set
    try:
        # Create a project data.
        # Note that this is a deprecated function and will be removed in the future.
        api_response: ProjectData = api_instance.create_data_in_project(
            project_id=str(project_id),
            create_data=CreateData(
                name=data_name,
                folderPath=parent_folder_path,
                dataType=data_type,
                folderId=None,
                formatCode=None,
            )
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->create_project_data: %s\n" % e)
        raise ApiException

    # Return the folder id
    return api_response


def create_file_in_project(
        project_id: Union[UUID4, str],
        file_path: Path,
) -> ProjectData:
    """
    Create a file in a project

    :param project_id: The project id to create the file in
    :param file_path: The path to the file

    :return: The newly created file

    :rtype: `ProjectData <https://umccr.github.io/libica/openapi/v3/docs/ProjectData/>`__

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import create_file_in_project
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_obj: ProjectData = create_file_in_project(
            project_id="abcd-1234-efab-5678",
            file_path=Path("/path/to/file.txt")
        )
    """

    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    parent_folder_path = str(file_path.absolute().parent) + "/"
    if parent_folder_path == "//":
        parent_folder_path = "/"

    # example passing only required values which don't have defaults set
    try:
        # Create a project data.
        api_response: ProjectData = api_instance.create_file(
            project_id=str(project_id),
            create_file_data=CreateFileData(
                name=file_path.name,
                folderPath=parent_folder_path,
                folderId=None,
                formatCode=None,
            )
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->create_file: %s\n" % e)
        raise ApiException

    # Return the folder id
    return api_response


def create_folder_in_project(
        project_id: Union[UUID4, str],
        folder_path: Path,
) -> ProjectData:
    """
    Create a folder in a project

    :param project_id:  The project ID
    :param folder_path:  The path to the folder to create

    :return: The newly created folder project data object
    :rtype: `ProjectData <https://umccr.github.io/libica/openapi/v3/docs/ProjectData/>`__

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import create_folder_in_project
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_obj: ProjectData = create_folder_in_project(
            project_id="abcd-1234-efab-5678",
            folder_path=Path("/path/to/folder/new/")
        )

    """

    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    parent_folder_path = str(folder_path.absolute().parent) + "/"
    if parent_folder_path == "//":
        parent_folder_path = "/"

    # example passing only required values which don't have defaults set
    try:
        # Create a project data.
        api_response: ProjectData = api_instance.create_folder(
            project_id=str(project_id),
            create_folder=CreateFolder(
                name=folder_path.name,
                folderPath=parent_folder_path,
                folderId=None
            )
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->create_folder: %s\n" % e)
        raise ApiException

    # Return the folder id
    return api_response


def get_project_data_folder_id_from_project_id_and_path(
        project_id: Union[UUID4, str],
        folder_path: Path,
        create_folder_if_not_found: bool = False
) -> str:
    """
    Given a project_id and a path, return the folder_id.

    Note that given the folder_path is a Path object (which do not end in /),
    we need to append a '/' to the end of the path before calling the API
    In the case that the folder_path is the root folder, we ensure that only a single '/' is provided.

    :param project_id:  The project id to search in
    :param folder_path:  The path to the folder
    :param create_folder_if_not_found:  If folder does not exist in project, do we want to create it?

    :return: The folder id

    :raises: NotADirectoryError, ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import get_project_folder_id_from_project_id_and_path

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        folder_id: str = get_project_folder_id_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            folder_path=Path("/path/to/folder/")
        )
    """
    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    parent_folder_path = str(folder_path.parent.absolute()) + "/"
    # Exception for when folder is in the top directory
    if parent_folder_path == '//':
        parent_folder_path = '/'

    # Add the folder name to the list of folder names to search on
    folder_name = [
        folder_path.name
    ]

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the list of project data.
        data_items: List[ProjectData] = api_instance.get_project_data_list(
            project_id=str(project_id),
            parent_folder_path=parent_folder_path,
            filename=folder_name,
            filename_match_mode="EXACT",
            file_path_match_mode="FULL_CASE_INSENSITIVE",
            type=FOLDER_DATA_TYPE
        ).items
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->get_project_data_list: %s\n" % e)
        raise ApiException

    # Get the folder id
    try:
        folder_id: ProjectData = next(
            filter(
                lambda data_iter: data_iter.data.details.path == str(folder_path) + "/",
                data_items
            )
        )
    except StopIteration:
        if create_folder_if_not_found:
            # Create the folder
            folder_id = create_folder_in_project(
                project_id=project_id,
                folder_path=folder_path
            )
        else:
            logger.error("Could not find folder id for folder: %s\n" % folder_path)
            raise NotADirectoryError

    return folder_id.data.id


def get_project_data_id_from_project_id_and_path(
        project_id: Union[UUID4, str],
        data_path: Path,
        data_type: DataType,
        create_data_if_not_found: bool = False
) -> str:
    """
    Given a project_id and a path, return the data_id, where DATA_TYPE is one of FILE or FOLDER
    Should call the underlying get_data_id_from_project_id_and_path function split by data type

    :param project_id: The project context the data exists in
    :param data_path: The path to the data in the project
    :param data_type: The data_type, one of "FILE", "FOLDER"
    :param create_data_if_not_found:

    :raises: FileNotFoundError, NotADirectoryError, ApiException

    :return: The data id

    :note:
      Use get_file_id_from_project_id_and_path or get_folder_id_from_project_id_and_path instead if data_type is known

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import get_project_data_id_from_project_id_and_path
        from wrapica.enums import DataType

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        # Get a folder
        data_id: str = get_project_data_id_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            data_path=Path("/path/to/folder/"),
            data_type="FOLDER"
        )

        # Get a file
        data_id: str = get_project_data_id_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            data_path=Path("/path/to/file.txt"),
            data_type="FILE"
        )
    """
    if data_type == FOLDER_DATA_TYPE:
        return get_project_data_folder_id_from_project_id_and_path(
            project_id=project_id,
            folder_path=data_path,
            create_folder_if_not_found=create_data_if_not_found
        )
    else:
        return get_project_data_file_id_from_project_id_and_path(
            project_id=project_id,
            file_path=data_path,
            create_file_if_not_found=create_data_if_not_found
        )


def get_project_data_obj_by_id(
        project_id: Union[UUID4, str],
        data_id: Union[UUID4, str]
) -> ProjectData:
    """
    Given a project_id and a data_id, return the data object

    :param project_id: The project id to search in
    :param data_id: The data id

    :return: The project data object
    :rtype: `ProjectData <https://umccr.github.io/libica/openapi/v3/docs/ProjectData/>`__

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import get_project_data_obj_by_id
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_obj: ProjectData = get_project_data_obj_by_id(
            project_id="abcd-1234-efab-5678",
            data_id="fil.abcdef1234567890"
        )
    """

    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the list of project data.
        data_obj: ProjectData = api_instance.get_project_data(
            project_id=str(project_id),
            data_id=str(data_id)
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->get_project_data_list: %s\n" % e)
        raise ApiException

    return data_obj


def get_project_data_obj_from_project_id_and_path(
        project_id: Union[UUID4, str],
        data_path: Path,
        data_type: DataType,
        create_data_if_not_found: bool = False
) -> ProjectData:
    """
    Given a project_id and a path, return the data object, where DATA_TYPE is one of FILE or FOLDER
    Will call the get_project_data_id and then call get_project_data_obj_by_id

    :param project_id: The project id to search in
    :param data_path: The path to the data in the project
    :param data_type: The data_type, one of "FILE", "FOLDER"
    :param create_data_if_not_found: If the data is not found, and create_data_if_not_found is True, create the data

    :return: The project data object
    :rtype: `ProjectData <https://umccr.github.io/libica/openapi/v3/docs/ProjectData/>`__

    :raises: FileNotFoundError, NotADirectoryError, ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import get_project_data_obj_from_project_id_and_path
        from wrapica.enums import DataType
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        # Get a folder project data object
        project_folder_data_obj: ProjectData = get_project_data_obj_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            data_path=Path("/path/to/folder/"),
            data_type="FOLDER"
        )

        # Get a file project data object
        project_file_data_obj: ProjectData = get_project_data_obj_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            data_path=Path("/path/to/file.txt"),
            data_type="FILE"
        )

        print(project_folder_data_obj.data.id)
        # fol.abcdef1234567890

        print(project_file_data_obj.data.id)
        # fil.abcdef1234567890
    """
    # Collect the data id, either fol.id or fil.id
    project_data_id = get_project_data_id_from_project_id_and_path(
        project_id=project_id,
        data_path=data_path,
        data_type=data_type,
        create_data_if_not_found=create_data_if_not_found
    )

    # Then collect the object itself
    return get_project_data_obj_by_id(
        project_id=project_id,
        data_id=project_data_id
    )


def get_project_data_path_by_id(
        project_id: Union[UUID4, str],
        data_id: Union[UUID4, str]
) -> Path:
    """
    Given a project id and data id, return the path of the data

    :param project_id: The project id to search in
    :param data_id: The data id

    :return: The path of the data
    :rtype: Path

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import get_project_data_path_by_id

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_path: Path = get_project_data_path_by_id(
            project_id="abcd-1234-efab-5678",
            data_id="fil.abcdef1234567890"
        )

        print(project_data_path)
        # /path/to/file.txt
    """
    project_data_path = get_project_data_obj_by_id(
        project_id=project_id,
        data_id=data_id
    ).data.details.path

    return Path(project_data_path)


def list_project_data_non_recursively(
        project_id: Union[UUID4, str],
        parent_folder_id: Optional[Union[UUID4, str]] = None,
        parent_folder_path: Optional[Path] = None,
        file_name: Optional[Union[str, List[str]]] = None,
        status: Optional[Union[ProjectDataStatusValuesType, List[ProjectDataStatusValuesType]]] = None,
        data_type: Optional[DataType] = None,
        creation_date_after: Optional[datetime] = None,
        creation_date_before: Optional[datetime] = None,
        status_date_after: Optional[datetime] = None,
        status_date_before: Optional[datetime] = None,
        sort: Optional[Union[ProjectDataSortParameterType, List[ProjectDataSortParameterType]]] = ""
) -> List[ProjectData]:
    """
    Given a project id and parent folder id or path,
    return a list of data objects that are directly under that folder

    :param project_id: The project id to search in
    :param parent_folder_path: The path to the parent folder (can use parent_folder_id instead)
    :param parent_folder_id: The parent folder id (can use parent_folder_path instead)
    :param file_name: The name of the file or directory to look for, can also be a list of names, may also use * as a wildcard
    :param status: The status of the data, one of ProjectDataStatusValues
    :param data_type: The type of the data, one of "FILE", "FOLDER"
    :param creation_date_after: Return only data created after this date
    :param creation_date_before: Return only data created before this date
    :param status_date_after: Return only data with status date after this date
    :param status_date_before: Return only data with status date before this date
    :param sort: The sort order, one or more of ProjectDataSortParameters (Use '-' prefix to sort in descending order)
      * timeCreated - Sort by time created
      * timeModified - Sort by time modified
      * name - Sort by name
      * path - Sort by path
      * fileSizeInBytes - Sort by file size in bytes
      * status - Sort by status
      * format - Sort by format
      * dataType - Sort by data type
      * willBeArchivedAt - Sort by when the data will be archived
      * willBeDeletedAt - Sort by when the data will be deleted

    :return: List of data objects
    :rtype: List[`ProjectData <https://umccr.github.io/libica/openapi/v3/docs/ProjectData/>`__]

    :raises: AssertionError, ApiException, ValueError

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.project_data import list_project_data_non_recursively
        from wrapica.libica_models import ProjectData, ProjectDataSortParameters
        from wrapica.enums import ProjectDataStatusValues, DataType

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_list: List[ProjectData] = list_project_data_non_recursively(
            project_id="abcd-1234-efab-5678",
            parent_folder_path=Path("/path/to/folder/"),
            file_name="file.txt",
            status=ProjectDataStatusValues.COMPLETED,
            data_type="FILE",
            creation_date_after=datetime(2021, 1, 1),
            creation_date_before=datetime(2021, 12, 31),
            status_date_after=datetime(2021, 1, 1),
            status_date_before=datetime(2021, 12, 31),
            sort=ProjectDataSortParameters.TIME_CREATED
        )

        for project_data in project_data_list:
            print(project_data.data.details.name)

    """
    # Check one of parent_folder_id and parent_folder_path is specified
    if parent_folder_id is None and parent_folder_path is None:
        logger.error("Must specify one of parent_folder_id and parent_folder_path")
        raise AssertionError
    elif parent_folder_id is not None and parent_folder_path is not None:
        logger.error("Must specify only one of parent_folder_id and parent_folder_path")
        raise AssertionError
    # Specify either parent_folder_id as a list parent_folder_path is just as a string
    if parent_folder_id is not None:
        parent_folder_id = [parent_folder_id]

    # Convert parent folder path to a string
    if parent_folder_path is not None:
        parent_folder_path = str(parent_folder_path.absolute()) + "/"
        if parent_folder_path == "//":
            parent_folder_path = "/"

    # Check file_name
    if isinstance(file_name, str):
        file_name = [file_name]

    # Check status
    if status is not None:
        if isinstance(status, ProjectDataStatusValuesType):
            status = [status]
        elif isinstance(status, str):
            status = [ProjectDataStatusValuesType(status)]
        status = list(
            map(
                lambda status_iter: status_iter.value,
                status
            )
        )

    # Check sort
    if sort == "":
        sort = None

    if sort is not None:
        if isinstance(sort, str) and not sort in ProjectDataSortParameterType.__args__:
            logger.error("Invalid sort parameter provided: %s" % sort)
            raise ValueError("Invalid sort parameter provided")
        elif isinstance(sort, List) and any(map(lambda sort_iter_: sort_iter_ not in ProjectDataSortParameterType.__args__, sort)):
            logger.error("Invalid sort parameter(s) provided: %s" % ", ".join(list(filter(
                lambda sort_iter_: sort_iter_ not in ProjectDataSortParameterType.__args__, sort
            ))))
            raise ValueError("Invalid sort parameter(s) provided")

        if isinstance(sort, str):
            sort = [sort]

        # Complete a comma join of the sort parameters
        sort = ", ".join(sort)

    # Collect api instance
    with ApiClient(get_icav2_configuration()) as api_client:
        api_instance = ProjectDataApi(api_client)

    # Set other parameters
    page_size = LIBICAV2_DEFAULT_PAGE_SIZE
    page_token = ""
    # We use page tokens if sort is None, otherwise we use page offsets
    if sort is not None:
        page_offset = 0
    else:
        page_offset = ""

    # Initialise data ids - we may need to extend the items multiple times
    data_obj_list: List[ProjectData] = []

    # Loop through the pages
    while True:
        # Attempt to collect all data ids
        try:
            # Retrieve the list of project data
            api_response = api_instance.get_project_data_list(
                **dict(
                    filter(
                        lambda x: x[1] is not None,
                        {
                            "status": status,
                            "type": data_type,
                            "project_id": project_id,
                            "parent_folder_id": parent_folder_id,
                            "parent_folder_path": parent_folder_path,
                            "page_size": str(page_size),
                            "page_offset": str(page_offset),
                            "page_token": page_token,
                            "filename": file_name,
                            "creation_date_after": creation_date_after,
                            "creation_date_before": creation_date_before,
                            "status_date_after": status_date_after,
                            "status_date_before": status_date_before,
                            "sort": sort
                        }.items()
                    )
                )
            )
        except ApiException as e:
            raise ValueError("Exception when calling ProjectDataApi->get_project_data_list: %s\n" % e)

        # Extend items list
        data_obj_list.extend(api_response.items)

        # Determine page iteration method by if we have a 'sort' parameter
        if sort is not None:
            # Check page offset and page size against total item count
            if page_offset + page_size >= api_response.total_item_count:
                break
            page_offset += page_size
        else:
            # Check if there is a next page
            if api_response.next_page_token is None or api_response.next_page_token == "":
                break
            page_token = api_response.next_page_token

    return data_obj_list


def find_project_data_recursively(
        project_id: Union[UUID4, str],
        parent_folder_id: Optional[Union[UUID4, str]] = None,
        parent_folder_path: Optional[Path] = None,
        name: Optional[str] = None,
        data_type: Optional[DataType] = None,
        min_depth: Optional[int] = None,
        max_depth: Optional[int] = None
) -> List[ProjectData]:
    """
    Given a project_id, a parent_folder_id, a data_name and a data_type, return a list of data objects
    This is a slow exercise and should only be used if the max_depth is low and the total number of items in the
    directory is very high

    :param project_id: The project id to search in
    :param parent_folder_id: The parent folder id (alternative to parent_folder_path)
    :param parent_folder_path: The path to the parent folder (alternative to parent_folder_id)
    :param name: The name of the file or directory to look for, may also use a regex pattern
    :param data_type: The type of the data, one of "FILE", "FOLDER"
    :param min_depth: The minimum depth to search
    :param max_depth: The maximum depth to search

    :return: List of data objects
    :rtype: List[`ProjectData <https://umccr.github.io/libica/openapi/v3/docs/ProjectData/>`__]

    :raises: ApiException, AssertionError, ValueError

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import find_project_data_recursively
        from wrapica.enums import DataType
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_list: List[ProjectData] = find_project_data_recursively(
            project_id="abcd-1234-efab-5678",
            parent_folder_id="fol.abcdef1234567890",
            name="file.txt",
            data_type="FILE",
            min_depth=1,
            max_depth=3
        )

        for project_data in project_data_list:
            print(project_data.data.details.name)

    """
    # Check one of parent_folder_id and parent_folder_path is specified
    if parent_folder_id is None and parent_folder_path is None:
        logger.error("Must specify one of parent_folder_id and parent_folder_path")
        raise AssertionError
    elif parent_folder_id is not None and parent_folder_path is not None:
        logger.error("Must specify only one of parent_folder_id and parent_folder_path")
        raise AssertionError

    # Matched data items thing we return
    matched_data_items: List[ProjectData] = []

    if name is not None and IS_REGEX_MATCH.match(name):
        name_recursive = name  # What we parse to this function recursively
        # If there are any * without a '.' before them, we need to add a '.' before them
        name = re.sub(r"(?<!\.)\*", ".*", name)

        name_regex_obj = re.compile(fr"{name}")
        name = None
    else:
        name_recursive = None
        name_regex_obj = None

    # Get top level items
    data_items: List[ProjectData] = list_project_data_non_recursively(
        project_id=project_id,
        parent_folder_id=parent_folder_id,
        parent_folder_path=parent_folder_path,
        data_type=data_type,
        file_name=name
    )

    # Check if we can pull out any items in the top directory
    if min_depth is None or min_depth <= 1:
        for data_item in data_items:
            # Check data type
            if data_type is not None and not data_item.data.details.data_type == data_type:
                continue
            # Check if we have regex name to match on
            if name_regex_obj is None:
                matched_data_items.append(data_item)
            elif name_regex_obj.fullmatch(data_item.data.details.name) is not None:
                matched_data_items.append(data_item)

    # Otherwise look recursively
    if max_depth is None or not max_depth <= 1:
        # Listing sub folders
        # If we didn't specify the datatype as FILE,
        # or a name / name regex, all the subfolders should be in the data items
        if not data_type == FILE_DATA_TYPE and name is None and name_regex_obj is None:
            subfolders = list(
                filter(
                    lambda x: x.data.details.data_type == FOLDER_DATA_TYPE,
                    data_items
                )
            )
        # Otherwise we will need to regather them
        else:
            subfolders = list_project_data_non_recursively(
                project_id=project_id,
                parent_folder_id=parent_folder_id,
                parent_folder_path=parent_folder_path,
                data_type=FOLDER_DATA_TYPE,
            )
        for subfolder in subfolders:
            matched_data_items.extend(
                find_project_data_recursively(
                    project_id=project_id,
                    parent_folder_id=subfolder.data.id,
                    name=name_recursive,
                    data_type=data_type,
                    min_depth=min_depth - 1 if min_depth is not None else None,
                    max_depth=max_depth - 1 if max_depth is not None else None
                )
            )

    return matched_data_items


def find_project_data_bulk(
        project_id: Union[UUID4, str],
        parent_folder_id: Optional[Union[UUID4, str]] = None,
        parent_folder_path: Optional[Path] = None,
        data_type: Optional[DataType] = None
) -> List[ProjectData]:
    """
    Given a project_id and a parent_folder_id, return a list of all data objects in the folder (recursively)

    :param project_id: The project id to search in
    :param parent_folder_id: The parent folder id (alternative to parent_folder_path)
    :param parent_folder_path: The path to the parent folder (alternative to parent_folder_id)
    :param data_type: The type of the data, one of "FILE", "FOLDER"

    :return: List of data objects
    :rtype: List[`ProjectData <https://umccr.github.io/libica/openapi/v3/docs/ProjectData/>`__]

    :raises: ApiException, AssertionError

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import find_project_data_bulk
        from wrapica.enums import DataType
        from wrapica.libica_models import ProjectData

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_list: List[ProjectData] = find_project_data_bulk(
            project_id="abcd-1234-efab-5678",
            parent_folder_id="fol.abcdef1234567890",
            data_type="FILE"
        )

        for project_data in project_data_list:
            print(project_data.data.details.name)
    """
    # Check one of parent_folder_id and parent_folder_path is specified
    if parent_folder_id is None and parent_folder_path is None:
        logger.error("Must specify one of parent_folder_id and parent_folder_path")
        raise AssertionError
    elif parent_folder_id is not None and parent_folder_path is not None:
        logger.error("Must specify only one of parent_folder_id and parent_folder_path")
        raise AssertionError

    # Get the parent folder path as a string
    if parent_folder_path is None:
        parent_folder_path = str(get_project_data_path_by_id(project_id, parent_folder_id)) + "/"
    else:
        parent_folder_path = str(parent_folder_path.absolute()) + "/"

    # Initialise
    data_ids: List[ProjectData] = []
    # Collect api instance
    with ApiClient(get_icav2_configuration()) as api_client:
        api_instance = ProjectDataApi(api_client)

    # Set other parameters
    page_size = LIBICAV2_DEFAULT_PAGE_SIZE
    page_token = ""

    # Iterate over all pages
    while True:
        # Attempt to collect all data ids
        try:
            # Retrieve the list of project data
            api_response = api_instance.get_project_data_list(
                project_id=str(project_id),
                file_path=[parent_folder_path],
                file_path_match_mode="STARTS_WITH_CASE_INSENSITIVE",
                page_size=str(page_size),
                page_token=page_token,
                type=data_type
            )

        except ApiException as e:
            logger.error("Exception when calling ProjectDataApi->get_project_data_list: %s\n" % e)
            raise ApiException

        # Extend items list
        data_ids.extend(api_response.items)

        # Check page offset and page size against total item count
        page_token = api_response.next_page_token

        if page_token == "":
            break

    return data_ids


def create_download_url(
        project_id: Union[UUID4, str],
        file_id: Union[UUID4, str]
) -> str:
    """
    Given a project_id and a data_id, create a presigned url for a file
    in project-pipeline, or view a file in project-data
    Create download URL

    :param project_id: The owning project id
    :param file_id: The id of the file

    :return: The download URL string
    :rtype: str

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import create_download_url

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        download_url: str = create_download_url(
            project_id="proj.abcdef1234567890",
            file_id="fil.abcdef1234567890"
        )

        print(download_url)
        # https://s3.amazonaws.com/umccr-illumina-prod/abcd-1234-efab-5678/abcdef1234567890

    """
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

        # example passing only required values which don't have defaults set
        try:
            # Retrieve a download URL for this data.
            api_response: Download = api_instance.create_download_url_for_data(
                project_id=str(project_id),
                file_id=str(file_id)
            )
        except ApiException as e:
            logger.error("Exception when calling ProjectDataApi->create_download_url_for_data: %s\n" % e)
            raise ApiException

    return api_response.url


def create_download_urls(
        project_id: Union[UUID4, str],
        folder_id: Union[UUID4, str],
        recursive: bool = False
) -> List[DataUrlWithPath]:
    """
    Given a project data folder return a list where each item is an object with the following attributes

    :param project_id: The owning project id
    :param folder_id:  The id of the folder
    :param recursive:  Whether to provide download urls recursively

    :return: List of download urls
    :rtype: List[`DataUrlWithPath <https://umccr.github.io/libica/openapi/v3/docs/DataUrlWithPathList/>`_]

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import (
            create_download_urls, get_project_folder_id_from_project_id_and_path
        )
        from wrapica.libica_models import ProjectData, DataUrlWithPath

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_folder_obj: ProjectData = get_project_folder_id_from_project_id_and_path(
            project_id="abcd-1234-efab-5678",
            folder_path=Path("/path/to/folder/")
        )

        download_urls: List[DataUrlWithPath] = create_download_urls(
            project_id="proj.abcdef1234567890",
            folder_id=project_folder_obj.data.id,
            recursive=True
        )

        for download_url in download_urls:
            print(download_url.url)
    """

    if recursive:
        project_data_list = find_project_data_bulk(
            project_id=project_id,
            parent_folder_id=folder_id,
            data_type=FILE_DATA_TYPE
        )
    else:
        project_data_list = list_project_data_non_recursively(
            project_id=project_id,
            parent_folder_id=folder_id,
            data_type=FILE_DATA_TYPE
        )

    # Set data paths
    data_id_paths_list = DataIdOrPathList(
        dataIds=list(
            map(
                lambda project_file_iter: project_file_iter.data.id,
                project_data_list
            )
        ),
        dataPaths=None
    )

    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v3+json"
        )

    # example passing only required values which don't have defaults set
    try:
        # Retrieve download URLs for the data.
        api_response = api_instance.create_download_urls_for_data(
            project_id=str(project_id),
            data_id_or_path_list=data_id_paths_list
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->create_download_urls_for_data: %s\n" % e)
        raise ApiException

    # Return items
    return api_response.items


def convert_icav2_uri_to_data_obj(
        data_uri: str,
        create_data_if_not_found: bool = False
) -> ProjectData:
    DeprecationWarning(
        "Please use convert_uri_to_project_data_obj or use "
        "convert_uri_to_data_obj from wrapica.data instead."
    )
    return convert_uri_to_project_data_obj(data_uri, create_data_if_not_found)


def convert_icav2_uri_to_project_data_obj(
        data_uri: str,
        create_data_if_not_found: bool = False
) -> ProjectData:
    DeprecationWarning(
        "Please use convert_uri_to_project_data_obj instead."
    )
    return convert_uri_to_project_data_obj(data_uri, create_data_if_not_found)


def convert_uri_to_project_data_obj(
        data_uri: str,
        create_data_if_not_found: bool = False
) -> ProjectData:
    """
    Given an ICAv2 URI, return a project data object

    :param data_uri: The icav2 uri to convert to a data object
    :param create_data_if_not_found:  If the data is not found, and create_data_if_not_found is True, create the data

    :return: libica v2 Project Data Object
    :rtype: `Project Data <https://umccr.github.io/libica/openapi/v3/docs/ProjectData/>`_

    :raises: ValueError, ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import convert_uri_to_project_data_obj, ProjectData

        project_data_object: ProjectData = convert_uri_to_project_data_obj(
            "icav2://project-name/path/to/data/"
        )

        print(project_data_object.data.id)
        # file.abcdef1234567890
    """
    # Import other functions locally to avoid circular imports
    from ...project import get_project_id_from_project_name

    data_uri_obj = urlparse(data_uri)

    # Set data type
    if data_uri_obj.path.endswith("/"):
        data_type = cast(DataType, FOLDER_DATA_TYPE)
    else:
        data_type = cast(DataType, FILE_DATA_TYPE)

    if cast(UriType, data_uri_obj.scheme) == ICAV2_URI_SCHEME:
        # Check if the project is in project id format
        if is_uuid_format(data_uri_obj.netloc):
            project_id = data_uri_obj.netloc
        else:
            project_id = get_project_id_from_project_name(data_uri_obj.netloc)
        data_path = Path(data_uri_obj.path)
    elif cast(UriType, data_uri_obj.scheme) == S3_URI_SCHEME:
        # If the uri is an s3 uri, we need to convert it to an icav2 uri
        project_id, data_path = unpack_uri(data_uri)
    else:
        logger.error(f"Could not convert uri to project data object, scheme {data_uri_obj.scheme} not recognised")
        raise ValueError

    # Return the data object
    return get_project_data_obj_from_project_id_and_path(
        project_id=project_id,
        data_path=Path(data_path),
        data_type=data_type,
        create_data_if_not_found=create_data_if_not_found
    )


def convert_project_data_obj_to_icav2_uri(
        project_data: ProjectData
) -> str:
    """
    Return the file object as an icav2:// uri

    :param project_data: The ProjectData object

    :return: The icav2:// uri string
    """
    DeprecationWarning(
        "Please use convert_project_data_obj_to_uri instead."
    )
    return convert_project_data_obj_to_uri(project_data, uri_type=ICAV2_URI_SCHEME)


def convert_project_data_obj_to_uri(
        project_data: ProjectData,
        uri_type: UriType = ICAV2_URI_SCHEME
) -> str:
    """
    Return the file object as an icav2:// uri

    :param project_data: The ProjectData object
    :param uri_type: One of UriType.ICAV2, UriType.S3

    :return: The icav2:// uri string
    """
    from ...storage_configuration import convert_project_data_obj_to_s3_uri
    if uri_type == ICAV2_URI_SCHEME:
        return str(
            urlunparse((
                uri_type,
                str(project_data.project_id),
                project_data.data.details.path.rstrip("/") + (
                    "/" if project_data.data.details.data_type == FOLDER_DATA_TYPE else ""),
                None, None, None
            ))
        )
    elif uri_type == S3_URI_SCHEME:
        return convert_project_data_obj_to_s3_uri(project_data_obj=project_data)
    else:
        logger.error(
            f"Uri type {uri_type} not recognised, please use one of UriType.ICAV2, UriType.S3"
        )
        raise ValueError


def convert_project_id_and_data_path_to_icav2_uri(
        project_id: Union[UUID4, str],
        data_path: Path,
        data_type: DataType
) -> str:
    DeprecationWarning(
        "Please use convert_project_id_and_data_path_to_uri instead."
    )
    return convert_project_id_and_data_path_to_uri(
        project_id=project_id,
        data_path=data_path,
        data_type=data_type
    )


def convert_project_id_and_data_path_to_uri(
        project_id: Union[UUID4, str],
        data_path: Path,
        data_type: DataType,
        uri_type: UriType = ICAV2_URI_SCHEME
) -> str:
    """
    Given a project_id and a data_id, return the icav2:// uri

    Unlike convert_project_data_obj_to_icav2_uri, this does not require the path to exist.

    If the data_type is "FOLDER", the path should end with a forward slash.

    :param project_id: The project id to search in
    :param data_path: The path to the data in the project
    :param data_type: The data_type, one of "FILE", "FOLDER"
    :param uri_type: One of UriType.ICAV2, UriType.S3

    :return: The icav2:// uri string
    :rtype: str

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_data import convert_project_id_and_data_path_to_uri
        from wrapica.enums import DataType

        icav2_uri: str = convert_project_id_and_data_path_to_icav2_uri(
            project_id="abcd-1234-efab-5678",
            data_path=Path("/path/to/folder/"),
            data_type="FOLDER"
        )
    """
    from ...storage_configuration import get_s3_key_prefix_by_project_id
    if uri_type == ICAV2_URI_SCHEME:
        return str(
            urlunparse((
                ICAV2_URI_SCHEME,
                str(project_id),
                str(data_path) + ("/" if data_type == FOLDER_DATA_TYPE else ""),
                None, None, None
            ))
        )
    elif uri_type == S3_URI_SCHEME:
        return str(
            urlunparse((
                S3_URI_SCHEME,
                str(project_id),
                str(
                    Path(get_s3_key_prefix_by_project_id(project_id)) / data_path
                ) + ("/" if data_type == FOLDER_DATA_TYPE else ""),
                None, None, None
            ))
        )
    else:
        logger.error("Error! Could not convert project id and data path to uri, uri scheme {uri_type} not recognised")
        raise ValueError


def unpack_icav2_uri(uri: str) -> Tuple[str, str]:
    DeprecationWarning(
        "Warning, please use unpack_uri instead."
    )
    return unpack_uri(uri)


def unpack_uri(uri: str) -> Tuple[str, str]:
    """
    Unpack an icav2 or s3 uri

    :param uri: The icav2 uri in the format of 'icav2://project_id/path/to/file.txt' or 'icav2://project_id/path/to/dir/'

    :return: Tuple with project_id and data_path

    :rtype: Tuple[str, str]

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import unpack_icav2_uri

        project_id, data_path = unpack_icav2_uri("icav2://project_id/path/to/dir")
    """
    # Get local imports
    from ...project import get_project_id_from_project_name
    from ...storage_configuration import unpack_s3_uri

    # Parse obj
    uri_obj = urlparse(uri)

    if cast(UriType, uri_obj.scheme) == ICAV2_URI_SCHEME:
        # Get project name or id
        project_name_or_id = uri_obj.netloc

        # Get data path
        data_path = uri_obj.path

        # Get project id
        if is_uuid_format(project_name_or_id):
            project_id = project_name_or_id
        else:
            project_id = get_project_id_from_project_name(project_name_or_id)

        return project_id, data_path
    elif cast(UriType, uri_obj.scheme) == S3_URI_SCHEME:
        return unpack_s3_uri(uri)
    else:
        raise ValueError(f"Could not unpack uri, scheme {uri_obj.scheme} not recognised")


def coerce_data_id_or_icav2_uri_to_project_data_obj(
        data_id_or_uri: str,
        create_data_if_not_found: bool = False
) -> ProjectData:
    DeprecationWarning(
        "Please use coerce_data_id_or_uri_to_project_data_obj instead "
    )
    return coerce_data_id_or_uri_to_project_data_obj(
        data_id_or_uri=data_id_or_uri,
        create_data_if_not_found=create_data_if_not_found
    )


def coerce_data_id_or_uri_to_project_data_obj(
        data_id_or_uri: str,
        create_data_if_not_found: bool = False
) -> ProjectData:
    """
    Given a data id or a uri, return a tuple of project id and data id

    :param data_id_or_uri:  The data id or uri
    :param create_data_if_not_found:  If uri_or_id is in uri format, and the data is not found, and create is True, create the data
    :return:
    """
    from ...data import get_project_data_obj_from_data_id
    if is_data_id_format(
            data_id=data_id_or_uri
    ):
        return get_project_data_obj_from_data_id(
            data_id=data_id_or_uri
        )
    return convert_uri_to_project_data_obj(
        data_uri=data_id_or_uri,
        create_data_if_not_found=create_data_if_not_found
    )


def coerce_data_id_icav2_uri_or_path_to_project_data_obj(
        data_id_path_or_uri: str,
        create_data_if_not_found: bool = False
) -> Optional[ProjectData]:
    DeprecationWarning(
        "Please use coerce_data_id_uri_or_path_to_project_data_obj or use "
        "coerce_data_id_uri_or_path_to_data_obj from wrapica.data instead."
    )
    return coerce_data_id_uri_or_path_to_project_data_obj(data_id_path_or_uri, create_data_if_not_found)


def coerce_data_id_uri_or_path_to_project_data_obj(
        data_id_path_or_uri: str,
        create_data_if_not_found: bool = False
) -> Optional[ProjectData]:

    """
    Coerce a data id, uri, path to a project data object

    :param data_id_path_or_uri:
    :param create_data_if_not_found:

    :return: A ProjectData object
    :rtype: `ProjectData <https://umccr.github.io/libica/openapi/v3/docs/ProjectData/>`__

    :raises: ValueError, ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import coerce_data_id_uri_or_path_to_project_data_obj

        project_data_obj = coerce_data_id_uri_or_path_to_project_data_obj(
            data_id_path_or_uri="icav2://project-name/path/to/data/"
        )
    """
    from ...project import get_project_id

    if is_data_id_format(data_id_path_or_uri):
        # Data ID, easy to convert across
        return get_project_data_obj_by_id(
            project_id=get_project_id(),
            data_id=data_id_path_or_uri
        )
    elif (
            is_uri_format(data_id_path_or_uri) and
            cast(UriType, urlparse(data_id_path_or_uri).scheme) in [ICAV2_URI_SCHEME, S3_URI_SCHEME]
    ):
        # ICAv2 URI, convert to data object
        return convert_uri_to_project_data_obj(
            data_uri=data_id_path_or_uri,
            create_data_if_not_found=create_data_if_not_found
        )
    else:
        # Data Path, convert to data object
        # Not as straight forward, need to first find this data, then convert to data object
        project_id = get_project_id()
        if Path(data_id_path_or_uri) == Path("/"):
            # There is no data id for the root directory (nor can we create one), we return none
            return None

        project_data_obj = get_project_data_obj_from_project_id_and_path(
            project_id=project_id,
            data_path=Path(data_id_path_or_uri),
            data_type=FOLDER_DATA_TYPE if data_id_path_or_uri.endswith("/") else FILE_DATA_TYPE,
            create_data_if_not_found=create_data_if_not_found
        )
        return project_data_obj


def get_aws_credentials_access_for_project_folder(
        project_id: Union[UUID4, str],
        folder_id: Optional[Union[UUID4, str]] = None,
        folder_path: Optional[Path] = None,
        read_only: Optional[bool] = None
) -> AwsTempCredentials:
    """
    Given a project_id and a folder_id or folder_path, collect the AWS Access Credentials for downloading this data.

    :param project_id: The project id of the data
    :param folder_id: The folder id (alternative to folder_path)
    :param folder_path: The folder path (alternative to folder_id)
    :param read_only: True to create read only credentials, otherwise defaults to read+write credentials

    :return: An object with the following attributes:
      * access_key
      * secret_key
      * session_token
      * region
      * bucket
      * object_prefix
      * server_side_encryption_algorithm
      * server_side_encryption_key

    :rtype: `AwsTempCredentials <https://umccr.github.io/libica/openapi/v3/docs/AwsTempCredentials/>`_

    :raises: AssertionError, ApiException, ValueError

    :Examples:

    .. code-block:: python
        :linenos:

        import subprocess
        from wrapica.project_data import get_aws_credentials_access_for_project_folder
        from wrapica.libica_models import AwsTempCredentials

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        aws_temp_credentials: AwsTempCredentials = get_aws_credentials_access_for_project_folder(
            project_id="proj.abcdef1234567890",
            folder_path=Path("/path/to/folder/")
        )

        local_path = Path("/path/to/local/folder/")

        subprocess.run(
          [
            "aws", "s3", "sync",
            # Can add sync parameters here like --dryrun or --exclude / --include
            "s3://{}/{}".format(aws_temp_credentials.bucket, aws_temp_credentials.object_prefix),
            str(local_path)
          ],
          env={
            "AWS_ACCESS_KEY_ID": aws_temp_credentials.access_key,
            "AWS_SECRET_ACCESS_KEY": aws_temp_credentials.secret_key,
            "AWS_SESSION_TOKEN": aws_temp_credentials.session_token,
            "AWS_REGION": aws_temp_credentials.region
          }
        )
    """
    # Check one of parent_folder_id and parent_folder_path is specified
    if folder_id is None and folder_path is None:
        logger.error("Must specify one of parent_folder_id and parent_folder_path")
        raise AssertionError
    elif folder_id is not None and folder_path is not None:
        logger.error("Must specify only one of parent_folder_id and parent_folder_path")
        raise AssertionError

    if folder_id is None:
        folder_id = get_project_data_folder_id_from_project_id_and_path(
            project_id=project_id,
            folder_path=folder_path
        )

    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v3+json"
        )

    # Create the temp credentials object
    create_temporary_credentials = CreateTemporaryCredentials(
        credentialsFormat="RCLONE",
        readOnlyCredentials=read_only
    )

    # example passing only required values which don't have defaults set
    try:
        # Retrieve temporary credentials for this data.
        api_response: TempCredentials = api_instance.create_temporary_credentials_for_data(
            project_id=str(project_id),
            data_id=str(folder_id),
            create_temporary_credentials=create_temporary_credentials
        )
    except ApiException as e:
        logger.warning("Exception when calling ProjectDataApi->create_temporary_credentials_for_data: %s\n" % e)
        raise ValueError

    return api_response.aws_temp_credentials


def is_folder_id_format(
        folder_id_str: str
) -> bool:
    """
    Does this string look like a folder id?

    :param folder_id_str: The string to check

    :return: True if the string looks like a folder id
    :rtype: bool

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import is_folder_id_format

        print(is_folder_id_format("fol.abcdef1234567890"))
        # True
    """
    return re.match("fol.[0-9a-f]{32}", folder_id_str) is not None


def is_file_id_format(
        file_id_str: str
) -> bool:
    """
    Does this string look like a folder id?

    :param file_id_str: The string to check

    :return: True if the string looks like a file id
    :rtype: bool

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import is_file_id_format

        print(is_file_id_format("fil.abcdef1234567890"))
        # True

    """
    return re.match("fil.[0-9a-f]{32}", file_id_str) is not None


def is_data_id_format(
        data_id: Union[UUID4, str]
) -> bool:
    """
    Check if data id is a data id

    :param data_id: The data id to check

    :return: True if the data id is a data id

    :rtype: bool

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import is_data_id

        print(is_data_id("fil.abcdef1234567890"))
        # True
    """
    return is_file_id_format(data_id) or is_folder_id_format(data_id)


def check_folder_exists(
        project_id: Union[UUID4, str],
        folder_path: Path
) -> bool:
    """
    Check folder path is a folder on icav2

    :param project_id:  The owning project id of the folder
    :param folder_path: The folder path

    :return: True if folder exists, otherwise false

    :rtype:  bool

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.project_data import check_folder_exists

        # Check if a folder exists
        print(check_folder_exists("abcdef1234567890", Path("/path/to/folder/")))

    """
    try:
        # Try to get data object from project id and path
        get_project_data_obj_from_project_id_and_path(project_id, folder_path, data_type=FOLDER_DATA_TYPE)
    except (ValueError, FileNotFoundError):
        return False
    else:
        return True


def check_file_exists(
        project_id: Union[UUID4, str],
        file_path: Path
) -> bool:
    """
    Check if a file exists in a project

    :param project_id:  The owning project id of the file
    :param file_path:   The file path

    :return: True if file exists, otherwise false
    :rtype:  bool

    :Examples:

    .. code-block:: python

        # Imports
        from wrapica.project_data import check_file_exists

        # Check if a file exists
        print(check_file_exists("abcdef1234567890", Path("/path/to/file.txt")))
    """
    try:
        # Try to get data object from project id and path
        get_project_data_obj_from_project_id_and_path(project_id, file_path, data_type=FILE_DATA_TYPE)
    except (ValueError, FileNotFoundError):
        return False
    else:
        return True


def check_uri_exists(
        data_uri: str
) -> bool:
    """

    Given a uri, check if the uri exists

    :param data_uri:

    :return: True if the uri exists, otherwise false

    :rtype: bool

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import check_uri_exists

        print(check_uri_exists("icav2://project-name/path/to/file.txt"))
    """
    if cast(UriType, urlparse(data_uri).scheme) in [ICAV2_URI_SCHEME, S3_URI_SCHEME]:
        project_id, data_path = unpack_uri(data_uri)
    else:
        raise ValueError(f"URI scheme '{urlparse(data_uri).scheme}' not supported")
    if data_path.endswith("/"):
        return check_folder_exists(project_id, Path(data_path))
    else:
        return check_file_exists(project_id, Path(data_path))


def presign_folder(
        project_id: Union[UUID4, str],
        folder_path: Optional[Path] = None,
        folder_id: Optional[Union[UUID4, str]] = None
) -> List[DataUrlWithPath]:
    """
    Presign a folder recursively

    Given a project_id and a folder_path or folder_id, return a list of presigned urls for the folder

    :param project_id:  The owning project id of the folder
    :param folder_path:  The folder path
    :param folder_id:  The folder id

    :return: List of presigned urls
    :rtype: List[`DataUrlWithPath <https://umccr.github.io/libica/openapi/v3/docs/DataUrlWithPathList/>`_]

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.project_data import presign_folder

        # Presign a folder
        presigned_urls_list = list(
            map(
                lambda data_uri_iter: data_uri_iter.url,
                presign_folder(
                    project_id="abcdef1234567890",
                    folder_path="/path/to/folder/"
                )
            )
        )
    """

    if folder_id is None:
        folder_id = get_project_data_id_from_project_id_and_path(
            project_id=project_id,
            data_path=folder_path,
            data_type=FOLDER_DATA_TYPE
        )

    return create_download_urls(
        project_id=project_id,
        folder_id=folder_id,
        recursive=True
    )


def presign_cwl_directory(
        project_id: Union[UUID4, str],
        data_id: Union[UUID4, str]
) -> List[
    Union[
        Dict[str, Union[Union[dict, str], Any]],
        Dict[str, Union[str, Any]]
    ]
]:
    """
    Given a CWL directory object, presign all files in the directory recursively, and return the list of presigned url

    :param project_id: The project id to search in
    :param data_id: The data id

    :return: The CWL input json Directory object where each file in the listing has a presigned url for a location attributes
    :rtype: List[Dict[str, Union[Union[dict, str], Any]]]

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import presign_cwl_directory

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id
        # Use wrapica.project_data.get_folder_id_from_project_id_and_path
        # If you need to convert a folder path to a folder id

        cwl_directory: List[Dict[str, Union[Union[dict, str], Any]]] = presign_cwl_directory(
            project_id="abcd-1234-efab-5678",
            data_id="fol.abcdef1234567890"
        )

        print(cwl_directory)
        # [
        #   {
        #     "class": "Directory",
        #     "basename": "folder",
        #     "listing": [
        #       {
        #         "class": "File",
        #         "basename": "file.txt",
        #         "location": "https://s3.amazonaws.com/umccr-illumina-prod/abcd-1234-efab-5678/abcdef1234567890"
        #       }
        #     ]
        #   }
        # ]
    """
    # Data ids
    cwl_item_objs = []

    # List items noncursively
    file_obj_list = list_project_data_non_recursively(
        project_id=project_id,
        parent_folder_id=data_id
    )

    # Collect file object list
    for file_item_obj in file_obj_list:
        data_type = cast(DataType, file_item_obj.data.details.data_type) # One of FILE | FOLDER
        data_id = file_item_obj.data.id
        basename = file_item_obj.data.details.name
        if data_type == FOLDER_DATA_TYPE:
            cwl_item_objs.append(
                {
                    "class": "Directory",
                    "basename": basename,
                    "listing": presign_cwl_directory(project_id, data_id)
                }
            )
        else:
            cwl_item_objs.append(
                {
                    "class": "File",
                    "basename": basename,
                    "location": create_download_url(project_id, data_id)
                }
            )

    return cwl_item_objs


def presign_cwl_directory_with_external_data_mounts(
        project_id: Union[UUID4, str],
        data_id: Union[UUID4, str]
) -> Tuple[
    # External data mounts
    List[AnalysisInputExternalData],
        # Dict listing
    List[Dict]
]:
    """
    Given a cwl directory with a listing attribute, presign all files in the directory recursively, and return the
    list of presigned url mount objects and the cwl directory listing object

    :param project_id: The project id to search in
    :param data_id: The data id

    :return: external_data_mounts, cwl_item_objs
    :rtype: Tuple[List[`AnalysisInputExternalData <https://umccr.github.io/libica/openapi/v3/docs/AnalysisInputExternalData/>`_], List[Dict]]

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import presign_cwl_directory_with_external_data_mounts

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id
        # Use wrapica.project_data.get_folder_id_from_project_id_and_path
        # If you need to convert a folder path to a folder id

        external_data_mounts, cwl_item_objs = presign_cwl_directory_with_external_data_mounts(
            project_id="abcd-1234-efab-5678",
            data_id="fol.abcdef1234567890"
        )

        print(external_data_mounts)
        # [
        #   {
        #     "url": "https://s3.amazonaws.com/umccr-illumina-prod/abcd-1234-efab-5678/abcdef1234567890",
        #     "type": "FILE",
        #     "mount_path": "abcd-1234-efab-5678/abcdef1234567890/file.txt"
        #   }
        # ]

        print(cwl_item_objs)
        # [
        #   {
        #     "class": "Directory",
        #     "basename": "folder",
        #     "listing": [
        #       {
        #         "class": "File",
        #         "basename": "file.txt",
        #         "location": "abcd-1234-efab-5678/abcdef1234567890/file.txt"
        #       }
        #     ]
        #   }
        # ]

    """
    # Data ids
    cwl_item_objs = []

    # External data mounts
    external_data_mounts = []

    # List items noncursively
    file_obj_list = list_project_data_non_recursively(
        project_id=project_id,
        parent_folder_id=data_id
    )

    # Collect file object list
    for file_item_obj in file_obj_list:
        data_type = cast(DataType, file_item_obj.data.details.data_type)  # One of FILE | FOLDER
        data_id = file_item_obj.data.id
        basename = file_item_obj.data.details.name
        if data_type == FOLDER_DATA_TYPE:
            external_data_mounts_new, listing = presign_cwl_directory_with_external_data_mounts(
                project_id,
                data_id
            )
            external_data_mounts.extend(external_data_mounts_new)
            cwl_item_objs.append(
                {
                    "class": "Directory",
                    "basename": basename,
                    "listing": listing
                }
            )
        else:
            # Generate presigned url
            presigned_url = create_download_url(project_id, data_id)

            # Generate mount path for file
            mount_path = str(
                Path(project_id) /
                Path(data_id) /
                Path(basename)
            )

            # Append the mount path and presigned url to the external data mounts list
            external_data_mounts.append(
                AnalysisInputExternalData(
                    url=presigned_url,
                    type="http",
                    mountPath=mount_path,
                    s3Details=None,
                    basespaceDetails=None,
                )
            )

            # Append the item to the cwl item object list
            cwl_item_objs.append(
                {
                    "class": "File",
                    "basename": basename,
                    "location": mount_path
                }
            )

    return external_data_mounts, cwl_item_objs


def read_icav2_file_contents(
        project_id: Union[UUID4, str],
        data_id: Union[UUID4, str],
        output_path: Optional[Union[Path, TextIOWrapper]] = None
) -> Optional[str]:
    """
    Write icav2 file contents to a path

    :param project_id: The project id
    :param data_id: The data id
    :param output_path: The output path to write the file contents to

    :return: The file contents as a string if output_path is None
    :rtype: Optional[str]

    :raises: NotADirectoryError, ApiException

    :Examples:

    .. code-block:: python

        :linenos:

        # Imports
        from wrapica.project_data import read_icav2_file_contents

        # Read icav2 file contents to a file
        with open("foo.txt", "w") as f:
            read_icav2_file_contents(
                project_id="abcd-1234-efab-5678",
                data_id="fil.abcdef1234567890",
                output_path=f
            )
    """
    if output_path is not None and isinstance(output_path, Path):
        # Ensure parent directory exists
        if not output_path.parent.exists():
            logger.error(f"Could not write to output path {output_path} as the parent directory does not exist")
            raise NotADirectoryError

    # Get the presigned url
    presigned_url = create_download_url(project_id, data_id)

    # Get the file contents with the requests package
    r = requests.get(presigned_url)

    if output_path is None:
        return r.content.decode()
    elif isinstance(output_path, Path):
        # Write the file contents to the output path
        with open(output_path, "wb") as f:
            f.write(r.content)
    else:
        # Write the file contents to the output path
        output_path.write(r.content.decode())

    return None


def read_icav2_file_contents_to_string(
        project_id: Union[UUID4, str],
        data_id: Union[UUID4, str]
) -> str:
    """

    Stream down the icav2 file contents and return as a string

    :param project_id: The project id
    :param data_id: The data id

    :return: The file contents as a string
    :rtype: str

    :raises: ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_data import read_icav2_file_contents_to_string

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        file_contents: str = read_icav2_file_contents_to_string(
            project_id="abcd-1234-efab-5678",
            data_id="fil.abcdef1234567890"
        )

        print(file_contents)
        # this is the file contents
    """

    with NamedTemporaryFile() as temp_file_h:
        read_icav2_file_contents(
            project_id=project_id,
            data_id=data_id,
            output_path=Path(temp_file_h.name)
        )

        with open(temp_file_h.name, "r") as f:
            return f.read()


def create_file_with_upload_url(
        project_id: Union[UUID4, str],
        folder_id: Union[UUID4, str],
        file_name: str
) -> str:
    """
    Create a new file in a project and return the upload URL

    :param project_id:  The owning project id of the file
    :param folder_id:  The folder id to create the file in
    :param file_name:  The name of the file to create
    :return:
    """

    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve an upload URL for this data.
        api_response: ProjectFileAndUploadUrl = api_instance.create_file_with_upload_url(
            project_id=str(project_id),
            create_file_and_upload_url=CreateFileAndUploadUrl(
                name=file_name,
                folderId=folder_id,
                folderPath=None,
                formatCode=None,
                fileType=None,
                hash=None,
            )
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->create_upload_url_for_data: %s\n" % e)
        raise ApiException("Exception when calling ProjectDataApi->create_upload_url_for_data: %s\n" % e) from e

    return api_response.upload_url


def get_project_data_upload_url(
        project_id: Union[UUID4, str],
        data_id: Union[UUID4, str]
) -> str:
    """
    Get upload url for project data object

    This can only be used for a file that has been created but not yet written to.

    :param project_id: The owning project id of the data
    :param data_id: The data id

    :return: The upload url
    :rtype: str

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.project_data import (
            get_project_data_upload_url,
            create_file_in_project
        )

        # Create a file in a project
        new_file_obj = create_file_in_project(
            project_id="abcd-1234-efab-5678",
            file_path=Path("/path/to/file.txt")
        )

        # Get the upload url for the new file
        upload_url = get_project_data_upload_url(
            project_id="abcd-1234-efab-5678",
            data_id=new_file_obj.data.id
        )
    """

    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve an upload URL for this data.
        api_response: Upload = api_instance.create_upload_url_for_data(
            project_id=str(project_id),
            data_id=str(data_id)
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->create_upload_url_for_data: %s\n" % e)
        raise ApiException

    return api_response.url


def write_icav2_file_contents(
        project_id: Union[UUID4, str],
        data_path: Path,
        file_stream_or_path: Union[Path, TextIOWrapper]
) -> str:
    """
    Write data to an icav2 file

    :param project_id: The owning project id of the file
    :param data_path: The file path
    :param file_stream_or_path: The file stream or path to write to

    :return: The new file id

    :rtype: str

    :raises: ValueError, ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.project_data import write_icav2_file_contents

        write_icav2_file_contents(
            project_id="abcd-1234-efab-5678",
            data_path=Path("/path/to/file.txt"),
            file_stream_or_path=Path("/path/to/local/file.txt")
        )
    """

    # Generate a new file in the project
    new_file_obj = create_file_in_project(
        project_id=project_id,
        file_path=data_path
    )

    # Get the upload url
    upload_url = get_project_data_upload_url(
        project_id=project_id,
        data_id=new_file_obj.data.id
    )

    if isinstance(file_stream_or_path, Path):
        with open(file_stream_or_path, "rb") as f:
            file_contents = f.read()
    else:
        file_contents = file_stream_or_path.read()

    # Upload file contents with the requests package
    requests.put(upload_url, data=file_contents)

    # Return the new file id
    return new_file_obj.data.id


def get_file_by_file_name_from_project_data_list(
        file_name: str,
        project_data_list: List[ProjectData]
) -> ProjectData:
    """
    Useful for collecting a file object from an analysis output object

    :param file_name: The name of the file to get
    :param project_data_list: The list of project data objects to search through

    :return: The file object
    :rtype: `ProjectData <https://umccr.github.io/libica/openapi/v3/docs/ProjectData/>`__

    :raises: ValueError

    :Examples:

    .. code-block:: python

        # Imports
        from wrapica.project_data import get_file_by_file_name_from_project_data_list

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        project_data_list: List[ProjectData] = find_project_data_bulk(
            project_id="abcd-1234-efab-5678",
            parent_folder_id="fol.abcdef1234567890",
            data_type="FILE"
        )

        file_obj: ProjectData = get_file_by_file_name_from_project_data_list(
            file_name="file.txt",
            project_data_list=project_data_list
        )
    """

    # Find the first file with this name
    try:
        return next(
            filter(
                lambda file_iter: (
                        file_iter.data.details.name == file_name and
                        file_iter.data.details.data_type == FILE_DATA_TYPE
                ),
                project_data_list
            )
        )
    except StopIteration:
        logger.error(f"Could not get file {file_name} from analysis output")
        raise ValueError


def project_data_copy_batch_handler(
        source_data_ids: List[str],
        destination_project_id: Union[UUID4, str],
        destination_folder_path: Path
) -> Job:
    """
    Copy a batch of files from one project to another

    :param source_data_ids: The list of source data ids
    :param destination_project_id: The destination project id
    :param destination_folder_path: The destination folder path

    :return: The job id for the project data copy batch
    :rtype: `Job <https://umccr.github.io/libica/openapi/v3/docs/Job/>`_

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_data import project_data_copy_batch_handler

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        job_id: str = project_data_copy_batch_handler(
            source_data_ids=[
                "fil.abcdef1234567890",
                "fil.abcdef1234567891"
            ],
            destination_project_id="abcd-1234-efab-5678",
            destination_folder_path=Path("/path/to/folder/")
        )
    """

    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataCopyBatchApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Copy a batch of project data.
        api_response: ProjectDataCopyBatch = api_instance.create_project_data_copy_batch(
            project_id=destination_project_id,
            create_project_data_copy_batch=CreateProjectDataCopyBatch(
                items=list(
                    map(
                        lambda source_data_id_iter: CreateProjectDataCopyBatchItem(
                            dataId=source_data_id_iter
                        ),
                        source_data_ids
                    )
                ),
                destinationFolderId=get_project_data_folder_id_from_project_id_and_path(
                    project_id=destination_project_id,
                    folder_path=destination_folder_path,
                    create_folder_if_not_found=True
                ),
                copyUserTags=True,
                copyTechnicalTags=True,
                copyInstrumentInfo=True,
                actionOnExist="SKIP"
            )
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->copy_project_data_batch: %s\n" % e)
        raise ApiException

    # Return the job ID for the project data copy batch
    return api_response.job


def delete_project_data(
        project_id: Union[UUID4, str],
        data_id: Union[UUID4, str]
):
    """
    Delete a project data item using the projectData:delete endpoint

    :param project_id: The project id the data belongs to
    :param data_id: The data id we want to delete

    :return: None
    :rtype: None

    :raises: ValueError, ApiException

    :Examples:

    .. code-block:: python

        from wrapica.project_data import delete_project_data

        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id

        delete_project_data(
            project_id="abcd-1234-efab-5678",
            data_id="fol.abcdef1234567890"
        )
    """
    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Force default headers for endpoints with a ':' in the name
        api_client.set_default_header(
            header_name="Content-Type",
            header_value="application/vnd.illumina.v3+json"
        )
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v3+json"
        )
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Schedule this data for deletion.
        api_instance.delete_data(
            project_id=str(project_id),
            data_id=str(data_id)
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->delete_data: %s\n" % e)
        raise ApiException


def move_project_data(
        dest_project_id: Union[UUID4, str],
        dest_folder_id: Union[UUID4, str],
        src_data_list: List[str]
) -> Job:
    """
    Move a list of data ids to a destination project
    :param dest_project_id:
    :param dest_folder_id:
    :param src_data_list:

    :return:

    :rtype: Job

    :raises: ApiException

    :Examples:

    .. code-block:: python

        from wrapica.project_data import move_data

        job = move_data(
            dest_project_id="abcd-1234-efab-5678",
            dest_folder_id="fol.abcdef1234567890",
            src_data_list=[
                "fil.abcdef1234567890",
                "fil.abcdef1234567891"
            ]
        )

    """

    # Create an instance of the API class
    with ApiClient(get_icav2_configuration()) as api_client:
        api_instance = ProjectDataMoveBatchApi(api_client)

    try:
        # Create a project data copy batch.
        api_response = api_instance.create_project_data_move_batch(
            project_id=str(dest_project_id),
            create_project_data_move_batch=CreateProjectDataMoveBatch(
                items=list(
                    map(
                        lambda src_data_iter: CreateProjectDataMoveBatchItem(
                            dataId=src_data_iter
                        ),
                        src_data_list
                    )
                ),
                destinationFolderId=dest_folder_id,
            )
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataMoveBatchApi->create_project_data_move_batch: %s\n" % e)
        raise ApiException("Exception when calling ProjectDataMoveBatchApi->create_project_data_move_batch: %s\n") from e

    # Get job from job id
    return api_response.job


def copy_project_data(
        dest_project_id: Union[UUID4, str],
        dest_folder_id: Union[UUID4, str],
        src_data_list: List[str]
) -> Job:
    """
    Copy a list of data ids to a destination project
    :param dest_project_id:
    :param dest_folder_id:
    :param src_data_list:

    :return:

    :rtype: Job

    :raises: ApiException

    :Examples:

    .. code-block:: python

        from wrapica.project_data import copy_project_data

        job = copy_project_data(
            dest_project_id="abcd-1234-efab-5678",
            dest_folder_id="fol.abcdef1234567890",
            src_data_list=[
                "fil.abcdef1234567890",
                "fil.abcdef1234567891"
            ]
        )

    """

    # Create an instance of the API class
    with ApiClient(get_icav2_configuration()) as api_client:
        api_instance = ProjectDataCopyBatchApi(api_client)

    try:
        # Create a project data copy batch.
        api_response = api_instance.create_project_data_copy_batch(
            project_id=str(dest_project_id),
            create_project_data_copy_batch=CreateProjectDataCopyBatch(
                items=list(
                    map(
                        lambda src_data_iter: CreateProjectDataCopyBatchItem(
                            dataId=src_data_iter
                        ),
                        src_data_list
                    )
                ),
                destinationFolderId=dest_folder_id,
                copyUserTags=False,
                copyTechnicalTags=False,
                copyInstrumentInfo=False,
                actionOnExist="SKIP"  # SKIP | OVERWRITE | FAIL
            )
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataCopyBatchApi->create_project_data_copy_batch: %s\n" % e)
        raise ApiException("Exception when calling ProjectDataCopyBatchApi->create_project_data_copy_batch: %s\n") from e

    # Get job from job id
    return api_response.job


def update_project_data_obj(
        project_id: Union[UUID4, str],
        data_id: Union[UUID4, str],
        project_data_obj: ProjectData
):
    """
    Given a project id, and data id, update with the project data object
    :param project_id:
    :param data_id:
    :param project_data_obj:
    :return:
    """
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectDataApi(api_client)

    try:
        # Update the project data object
        api_response: ProjectData = api_instance.update_project_data(
            project_id=str(project_id),
            data_id=str(data_id),
            project_data=project_data_obj
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectDataApi->update_project_data: %s\n" % e)
        raise e

    return api_response


def add_tag_to_data_object(
        project_id: Union[UUID4, str],
        data_id: Union[UUID4, str],
        tag: str,
        tag_type: DataTagType
) -> ProjectData:
    """
    Add tags to a data object

    :param project_id:
    :param data_id:
    :param tag:
    :param tag_type:
    :return:

    :return:

    :raises: ApiException

    :Examples:

    .. code-block:: python

        add_tag_to_data_obj(
          project_id,
          data_id,
          "to_be_archived"
        )
    """
    # Get the existing object
    project_data_obj = get_project_data_obj_by_id(
        project_id=project_id,
        data_id=data_id
    )

    # Get existing tags
    if tag_type == "technical_tag":
        project_data_obj.data.details.tags.technical_tags.append(tag)
    elif tag_type == "user_tag":
        project_data_obj.data.details.tags.user_tags.append(tag)
    elif tag_type == "connector_tag":
        project_data_obj.data.details.tags.connector_tags.append(tag)
    elif tag_type == "run_in_tag":
        project_data_obj.data.details.tags.run_in_tags.append(tag)
    elif tag_type == "run_out_tag":
        project_data_obj.data.details.tags.run_out_tags.append(tag)
    elif tag_type == "reference_tag":
        project_data_obj.data.details.tags.reference_tags.append(tag)
    else:
        raise ValueError("Tag type not recognised")

    # Update the analysis object
    project_data_obj = update_project_data_obj(
        project_id=project_id,
        data_id=data_id,
        project_data_obj=project_data_obj
    )

    return project_data_obj



# Model imports
from libica.openapi.v3.models import Region

# Local function imports
from .functions.region_functions import (
    get_regions,
    get_region_obj_from_region_id,
    get_region_obj_from_city_name,
    coerce_region_id_or_city_name_to_region_obj,
    coerce_region_id_or_city_name_to_region_id,
    get_region_obj_from_project_id,
    get_region_from_bundle_id,
    set_default_region,
    get_default_region
)


__all__ = [
    # Models
    'Region',
    # Functions
    'get_regions',
    'get_region_obj_from_region_id',
    'get_region_obj_from_city_name',
    'coerce_region_id_or_city_name_to_region_obj',
    'coerce_region_id_or_city_name_to_region_id',
    'get_region_obj_from_project_id',
    'get_region_from_bundle_id',
    'set_default_region',
    'get_default_region',
]



#!/usr/bin/env python

# Model imports
from libica.openapi.v3.models import (
    Job
)

# Import everything
from .functions.job_functions import (
    # Job functions
    get_job,
    wait_for_job_completion
)

__all__ = [
    # Models
    'Job',
    # Functions
    'get_job',
    'wait_for_job_completion'
]



#!/usr/bin/env python3

"""
CWL helpers

Handy for generating input templates, i.e from a cwl object,
Generate an input yaml template for the cwl object

Can also generate the overrides template from a cwl object
"""
# External imports
from copy import deepcopy
from pathlib import Path
from typing import List, Dict, Union, Optional
import re
from urllib.parse import urlparse
from ruamel.yaml import YAML, CommentedMap

# CWL Util imports
from cwl_utils.parser.cwl_v1_2 import shortname, RecordSchema, WorkflowStep

# Local imports
from .logger import get_logger
from .cwl_typing_helpers import (
    WorkflowInputParameterType,
    InputEnumSchemaType,
    RecordSchemaType,
    InputArraySchemaType,
    InputRecordSchemaType
)

# Get logger
logger = get_logger()


class CWLSchema:
    """
    Missing component of cwlutils
    """

    def __init__(self, cwl_obj, file_path):
        self.cwl_obj: RecordSchemaType = cwl_obj
        self.cwl_file_path: Path = file_path

        # Confirm type is record
        if not self.cwl_obj.type_.get("type") == "record":
            logger.error("Expected record type")

    def get_input_from_str_type(self, workflow_input: Dict) -> Optional[Union[Dict, str, List]]:
        if workflow_input.get("type").endswith("[]"):
            new_workflow_input = deepcopy(workflow_input)
            new_workflow_input["type"] = re.sub(r"\[]$", "", workflow_input.get("type"))
            return [
                self.get_input_from_str_type(new_workflow_input)
            ]
        if workflow_input.get("type").rstrip("?") == "Directory":
            return {
                "class": "Directory",
                "location": "icav2://project_id/path/to/dir/"
            }
        elif workflow_input.get("type").rstrip("?") == "File":
            return {
                "class": "File",
                "location": "icav2://project_id/path/to/file"
            }
        elif workflow_input.get("type").rstrip("?") == "boolean":
            return workflow_input.get("default") if workflow_input.get("default") is not None else False
        elif workflow_input.get("type").rstrip("?") == "int":
            return workflow_input.get("default") if workflow_input.get("default") is not None else "string"
        elif workflow_input.get("type").rstrip("?") == "string":
            return workflow_input.get("default") if workflow_input.get("default") is not None else "string"
        logger.warning("Don't know what to do here with {workflow_input.get('type')}")
        return None

    def get_input_from_array_type(self, workflow_input: Dict) -> Union[Dict, str, List]:
        """
        Likely first element of type is null
        :param workflow_input:
        :return:
        """
        workflow_input_new = deepcopy(workflow_input)
        if workflow_input.get("type")[0] == "null":
            workflow_input_new["type"] = workflow_input.get("type")[1]
        else:
            workflow_input_new["type"] = workflow_input.get("type")[0]

        if isinstance(workflow_input_new.get("type"), Dict):
            return self.get_input_from_dict_type(workflow_input_new)
        elif isinstance(workflow_input_new.get("type"), List):
            return self.get_input_from_array_type(workflow_input_new)
        elif isinstance(workflow_input_new.get("type"), str):
            return self.get_input_from_str_type(workflow_input_new)
        else:
            logger.error(f"Unsure what to do with type {type(workflow_input_new.get('type'))}")
            raise NotImplementedError

    def get_input_from_record_type(self, workflow_input: Dict) -> Union[Dict]:
        """
        Very similar to schema base command
        :param workflow_input:
        :return:
        """
        workflow_inputs = {}
        for field_key, field_dict in workflow_input.get("fields").items():
            if isinstance(field_dict.get("type"), Dict):
                workflow_inputs.update(
                    {
                        field_key: self.get_input_from_dict_type(field_dict)
                    }
                )
            elif isinstance(field_dict.get("type"), List):
                workflow_inputs.update(
                    {
                        field_key: self.get_input_from_array_type(field_dict)
                    }
                )
            elif isinstance(field_dict.get("type"), str):
                workflow_inputs.update(
                    {
                        field_key: self.get_input_from_str_type(field_dict)
                    }
                )
            else:
                logger.warning(f"Don't know what to do with type {type(field_dict.get('type'))} for key {field_key}")
        return workflow_inputs

    def get_input_from_dict_type(self, workflow_input: Dict) -> Optional[Union[Dict, List]]:
        """
        Dict type
        :param workflow_input:
        :return:
        """
        if "type" in workflow_input.get("type").keys() and workflow_input.get("type").get("type") == "record":
            return self.get_input_from_record_type(workflow_input.get("type"))
        if "type" in workflow_input.get("type").keys() and workflow_input.get("type").get("type") == "array":
            if isinstance(workflow_input.get("type").get("items"), str):
                return [
                    self.get_input_from_str_type(
                        {
                            "type": workflow_input.get("type").get("items")
                        }
                    )
                ]
            elif isinstance(workflow_input.get("type").get("items"), Dict):
                # We have an import
                return self.get_input_from_dict_type(
                    {
                        "type": workflow_input.get("type").get("items")
                    }
                )
        if "$import" in workflow_input.get("type").keys():
            schema_path = self.cwl_file_path.parent.joinpath(
                workflow_input.get("type").get("$import").split("#", 1)[0]
            ).resolve()
            return CWLSchema.load_schema_from_uri(schema_path.as_uri()).get_inputs_template()
        logger.warning(f"Got an unexpected type {workflow_input.get("type")}")
        return None


    def get_inputs_template(self) -> Dict:
        """
        Return get inputs from dict
        :return:
        """
        workflow_inputs = {}
        for field_key, field_dict in self.cwl_obj.type_.get("fields").items():
            if isinstance(field_dict.get("type"), Dict):
                workflow_inputs.update(
                    {
                        field_key: self.get_input_from_dict_type(field_dict)
                    }
                )
            elif isinstance(field_dict.get("type"), List):
                workflow_inputs.update(
                    {
                        field_key: self.get_input_from_array_type(field_dict)
                    }
                )
            elif isinstance(field_dict.get("type"), str):
                workflow_inputs.update(
                    {
                        field_key: self.get_input_from_str_type(field_dict)
                    }
                )
            else:
                logger.warning(f"Don't know what to do with type {type(field_dict.get('type'))} for key {field_key}")
        return workflow_inputs

    @classmethod
    def load_schema_from_uri(cls, uri_input):
        file_path: Path = Path(urlparse(uri_input).path)

        yaml_handler = YAML()

        with open(file_path, "r") as schema_h:
            schema_obj = yaml_handler.load(schema_h)

        return cls(RecordSchema(schema_obj), file_path)


def get_workflow_input_type(workflow_input: WorkflowInputParameterType):
    if isinstance(workflow_input.type_, str):
        return get_workflow_input_type_from_str_type(workflow_input)
    elif isinstance(workflow_input.type_, InputEnumSchemaType):
        return get_workflow_input_type_from_enum_schema(workflow_input)
    elif isinstance(workflow_input.type_, InputArraySchemaType):
        return get_workflow_input_type_from_array_schema(workflow_input)
    elif isinstance(workflow_input.type_, InputRecordSchemaType):
        return get_workflow_input_type_from_record_schema(workflow_input)
    elif isinstance(workflow_input.type_, List):
        return get_workflow_input_type_from_array_type(workflow_input)

    logger.warning(f"Don't know what to do here with {type(workflow_input.type_)}")
    return None


def get_workflow_input_type_from_str_type(workflow_input: WorkflowInputParameterType):
    """
    Workflow input type is a string type
    :param workflow_input:
    :return: A list with the following attributes
      {

      }
    """
    if workflow_input.type_.startswith("file://"):
        # This is a schema!
        return CWLSchema.load_schema_from_uri(workflow_input.type_).get_inputs_template()
    if workflow_input.type_ == "Directory":
        return {
            "class": "Directory",
            "location": "icav2://project_id/path/to/dir/"
        }
    elif workflow_input.type_ == "File":
        return {
            "class": "File",
            "location": "icav2://project_id/path/to/file"
        }
    elif workflow_input.type_ == "boolean":
        return workflow_input.default if workflow_input.default is not None else False
    elif workflow_input.type_ == "int":
        return workflow_input.default if workflow_input.default is not None else "string"
    elif workflow_input.type_ == "string":
        return workflow_input.default if workflow_input.default is not None else "string"

    logger.warning(f"Don't know what to do here with {workflow_input.type_}")
    return None


def get_workflow_input_type_from_array_type(workflow_input: WorkflowInputParameterType):
    """
    Workflow input is type list -
    likely that the first input is 'null'
    :param workflow_input:
    :return:
    """
    if not workflow_input.type_[0] == "null":
        logger.error("Unsure what to do with input of type list where first element is not null")
        raise ValueError
    workflow_input_new = deepcopy(workflow_input)
    workflow_input_new.type_ = workflow_input.type_[1]
    return get_workflow_input_type(workflow_input_new)


def get_workflow_input_type_from_enum_schema(workflow_input: WorkflowInputParameterType):
    """
    Workflow input type is a enum type
    :param workflow_input:
    :return:
    """
    workflow_type: InputEnumSchemaType = workflow_input.type_
    return shortname(workflow_type.symbols[0])


def get_workflow_input_type_from_array_schema(workflow_input: WorkflowInputParameterType):
    """
    Workflow input type is an array schema
    items attribute may be a file uri
    :param workflow_input:
    :return:
    """
    if workflow_input.type_.items == "string":
        return [
            "string" if workflow_input.default is None else workflow_input.default
        ],
    elif workflow_input.type_.items == "int":
        return [
            1 if workflow_input.default is None else workflow_input.default
        ],
    elif workflow_input.type_.items == "boolean":
        return [
            False if workflow_input.default is None else workflow_input.default
        ]
    elif workflow_input.type_.items == "File":
        return [
            {
                "class": "File",
                "location": "icav2://project_id/path/to/file"
            }
        ]
    elif workflow_input.type_.items == "Directory":
        return [
            {
                "class": "Directory",
                "location": "icav2://project_id/path/to/dir/"
            }
        ]
    return [
        CWLSchema.load_schema_from_uri(workflow_input.type_.items).get_inputs_template()
    ]


def get_workflow_input_type_from_record_schema(workflow_input: WorkflowInputParameterType):
    raise NotImplementedError


def get_workflow_input_is_optional(workflow_input: WorkflowInputParameterType):
    if isinstance(workflow_input.type_, List) and workflow_input.type_[0] == 'null':
        return True
    elif getattr(workflow_input, 'default') is not None:
        return True
    return False


def create_template_from_workflow_inputs(
        workflow_inputs: List[WorkflowInputParameterType]
) -> CommentedMap:
    """
    List inputs by template
    :param workflow_inputs:

    :return: A commented map of the inputs
    :rtype: CommentedMap
    """
    input_type_map = CommentedMap()

    for workflow_input in workflow_inputs:
        input_type_map.update(
            {
                shortname(workflow_input.id): get_workflow_input_type(workflow_input)
            }
        )

        # Add label
        before_comments = [
            f"{workflow_input.label} ({'Required' if not get_workflow_input_is_optional(workflow_input) else 'Optional'})"
        ]

        # Add default if it exists
        if workflow_input.default is not None:
            before_comments.append(f"Default value: {workflow_input.default}")

        # Add docs
        before_comments.append(f"Docs: {workflow_input.doc}")

        # Write comment to commented map
        input_type_map.yaml_set_comment_before_after_key(
            key=shortname(workflow_input.id),
            before='\n' + '\n'.join(before_comments),
            after="\n"
        )

    return input_type_map


def get_overrides_from_workflow_steps(workflow_steps: List[WorkflowStep]) -> Dict:
    """
    Get overrides from workflow steps
    :param workflow_steps:
    :return:
    """
    overrides = CommentedMap()

    for step in workflow_steps:
        overrides.update(
            {
                shortname(step.id): ""
            }
        )
    return overrides



#!/usr/bin/env python3

# Standard library imports
from typing import Union
from pydantic import UUID4

# Libica Api imports
from libica.openapi.v3.api.user_api import UserApi
from libica.openapi.v3 import ApiClient, ApiException

# Libica model imports
from libica.openapi.v3.models import (
    User,
    UserList
)

# Local imports
from ...utils.globals import ICAV2_ACCESS_TOKEN_AUDIENCE
from ...utils.logger import get_logger
from ...utils.configuration import get_icav2_configuration, get_jwt_token_obj
from ...utils.miscell import is_uuid_format

# Get logger
logger = get_logger()


def get_user_obj_from_user_id(
        user_id: Union[UUID4, str]
) -> User:
    """
    Get the user object from the user id

    :param user_id:

    :return: `User <https://umccr.github.io/libica/openapi/v3/docs/User/>`_
    """
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = UserApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve a user.
        api_response: User = api_instance.get_user(user_id)
    except ApiException as e:
        logger.error("Exception when calling UserApi->get_user: %s\n" % e)
        raise ApiException

    return api_response


def get_user_name_from_user_id(
        user_id: Union[UUID4, str]
) -> str:
    """
    Get the user name from the user id

    :param user_id: The user id

    :return: The user name
    :rtype: str
    """
    return get_user_obj_from_user_id(user_id).username


def get_user_obj_from_user_name(user_name: str) -> User:
    """
    Get the user object from the user name
    Lists through all user objects, when we have both user id and user name, use get_user_from_user_id instead

    :param user_name:

    :return: The user object
    :rtype: `User <https://umccr.github.io/libica/openapi/v3/docs/User/>`_

    :raises ApiException: If an error occurs when collecting the users
    :raises ValueError: If the user name is not found
    """
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = UserApi(api_client)

    # Get the user
    try:
        # Retrieve a user.
        api_response: UserList = api_instance.get_users()
    except ApiException as e:
        raise ApiException("Exception when calling UserApi->get_user: %s\n" % e)

    try:
        return next(
            filter(
                lambda x: x.firstname + " " + x.lastname == user_name,
                api_response.items
            )
        )
    except StopIteration:
        logger.error(f"Could not find user name '{user_name}'")
        raise ValueError


def coerce_user_id_or_name_to_user_obj(user_id_or_user_name: str) -> User:
    """
    Given either a user id or name, coerce to user object

    :param user_id_or_user_name:

    :return: The user object
    :rtype: `User <https://umccr.github.io/libica/openapi/v3/docs/User/>`_

    :raises ValueError: If the user name is not found
    :raises ApiException: If an error occurs when collecting the users

    :Example:

    .. code-block:: python
        :linenos:

        my_user: User = coerce_user_id_or_user_name_to_user_obj("user_id")
    """
    if is_uuid_format(user_id_or_user_name):
        return get_user_obj_from_user_id(user_id_or_user_name)

    return get_user_obj_from_user_name(user_id_or_user_name)


def coerce_user_id_or_name_to_user_id(user_id_or_user_name: str) -> str:
    """
    Given either a user id or name, coerce to user id

    :param user_id_or_user_name:

    :return: The user id
    """
    if is_uuid_format(user_id_or_user_name):
        return user_id_or_user_name

    return str(get_user_obj_from_user_name(user_id_or_user_name).id)


def get_user_id_from_configuration() -> str:
    """
    Use jwt to get username from access token

    :return: The user id
    :rtype: str
    """
    return get_jwt_token_obj(get_icav2_configuration().access_token, ICAV2_ACCESS_TOKEN_AUDIENCE).get("sub")


def get_tenant_id_for_user():
    """
    Get user and return tenant id from user

    :return: The tenant id
    :rtype: str
    """
    user_id = get_user_id_from_configuration()

    return get_user_obj_from_user_id(user_id).tenant_id



#!/usr/bin/env python3

# Standard imports
from typing import List, Optional, Union
from pydantic import UUID4

# Libica Api imports
from libica.openapi.v3 import ApiClient, ApiException
from libica.openapi.v3.api.region_api import RegionApi

# Libica model imports
from libica.openapi.v3.models import Region

# Local imports
from ...utils.logger import get_logger
from ...utils.configuration import get_icav2_configuration
from ...utils.miscell import is_uuid_format

# Get logger
logger = get_logger()

DEFAULT_REGION: Optional[Region] = None


def get_regions() -> List[Region]:
    """
    Return a list of regions

    :return: The list of regions available to the user in this tenant
    :rtype: List[`Region <https://umccr.github.io/libica/openapi/v3/docs/Region/>`_]
    :raises ApiException: If an error occurs while retrieving the regions

    :Example:

    .. code-block:: python
        :linenos:

        from wrapica.region import get_regions

        regions = get_regions()

        if len(regions) == 0:
            print("No regions found")
        else:
            print(f"Found {len(regions)} region(s)")
            for region in regions:
                print(f"Region ID: {region.id}, City Name: {region.city_name}")
    """
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = RegionApi(api_client)

    # example, this endpoint has no required or optional parameters
    try:
        # Retrieve a list of regions. Only the regions the user has access to through his/her entitlements are returned.
        api_response = api_instance.get_regions()
    except ApiException as e:
        logger.error("Exception when calling RegionApi->get_regions: %s\n" % e)
        raise ApiException

    return api_response.items


def get_region_obj_from_region_id(
        region_id: Union[UUID4, str]
) -> Region:
    """
    Get region object from the region id

    :param region_id:  The region ID
    :return: The region object
    :rtype: `Region <https://umccr.github.io/libica/openapi/v3/docs/Region/>`

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.region import get_region_obj_from_region_id

        region_id = "region-1234"
        region = get_region_obj_from_region_id(region_id)

        print(f"Region ID: {region.id}, City Name: {region.city_name}")
    """
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = RegionApi(api_client)

    # example, this endpoint has no required or optional parameters
    try:
        # Get region
        api_response = api_instance.get_region(region_id)
    except ApiException as e:
        logger.error("Exception when calling RegionApi->get_regions: %s\n" % e)
        raise ApiException

    return api_response


def get_region_obj_from_city_name(city_name: str) -> Region:
    """
    Get the region id from the city name

    :param city_name: The city name
    :return: The region object

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.region import get_region_obj_from_city_name

        city_name = "Sydney"
        region = get_region_obj_from_city_name(city_name)

        print(f"Region ID: {region.id}, City Name: {region.city_name}")
    """
    try:
        return next(
            filter(
                lambda region_iter: region_iter.city_name == city_name,
                get_regions()
            )
        )
    except StopIteration:
        logger.error(f"Could not get region object from city name {city_name}")
        raise StopIteration


def coerce_region_id_or_city_name_to_region_obj(region_id_or_city_name: str) -> Region:
    """
    Given either a region id or a region city name, coerce to region object

    :param region_id_or_city_name: The region id or city name
    :return: The region object
    :rtype: `Region <https://umccr.github.io/libica/openapi/v3/docs/Region/>`_
    """
    if is_uuid_format(region_id_or_city_name):
        return get_region_obj_from_region_id(region_id_or_city_name)
    return get_region_obj_from_city_name(region_id_or_city_name)


def coerce_region_id_or_city_name_to_region_id(region_id_or_city_name: str) -> str:
    """
    Given either a region id or a region city name, coerce to region id

    :param region_id_or_city_name:

    :return: The region id
    :rtype: str
    """
    if is_uuid_format(region_id_or_city_name):
        return region_id_or_city_name
    return str(get_region_obj_from_city_name(region_id_or_city_name).id)


def get_region_obj_from_project_id(
        project_id: Union[UUID4, str]
) -> Region:
    """
    Collect the region object from the project id

    :param project_id: The project ID
    :return: The region object
    :rtype: `Region <https://umccr.github.io/libica/openapi/v3/docs/Region/>`_

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.region import get_region_obj_from_project_id

        project_id = "project-1234"
        region = get_region_obj_from_project_id(project_id)
    """
    from ...project import get_project_obj_from_project_id
    return get_project_obj_from_project_id(project_id).region


def get_region_from_bundle_id(
        bundle_id: Union[UUID4, str]
) -> Region:
    """
    Get the region object from the bundle id

    :param bundle_id: The bundle ID

    :return: The region object
    :rtype: `Region <https://umccr.github.io/libica/openapi/v3/docs/Region/>`_

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.region import get_region_from_bundle_id

        bundle_id = "bundle-1234"
        region = get_region_from_bundle_id(bundle_id)
    """
    from ...bundle import get_bundle_obj_from_bundle_id
    bundle_obj = get_bundle_obj_from_bundle_id(bundle_id)
    return bundle_obj.region


def set_default_region():
    """
    Set the default region
    Assumes only one region is available to the user
    This is used by get_default_region,
    region is stored as a global python variable and is available for the rest of the session.

    """

    global DEFAULT_REGION

    regions = get_regions()

    if len(regions) == 0:
        raise Exception("No regions found, could not set default region")

    if not len(regions) == 1:
        raise Exception("Multiple regions found, cannot set default region")

    DEFAULT_REGION = regions[0]


def get_default_region() -> Region:
    """
    Get the default region, if no region is set, invocate set_default_region and assign global variable
    Used by bundle and project functions to get the default region when no region is provided

    :return: The default region object
    :rtype: `Region <https://umccr.github.io/libica/openapi/v3/docs/Region/>`

    """
    if DEFAULT_REGION is None:
        set_default_region()
    return DEFAULT_REGION




#!/usr/bin/env python3

"""
All the literals!
"""

# Standard library imports
from datetime import datetime
from typing import Literal, TypedDict, Optional

BundleStatusType = Literal[
    "DRAFT",
    "RELEASED",
    "DEPRECATED",
]


DataType = Literal[
    "FOLDER",
    "FILE",
]

AnalysisStorageSizeType = Literal[
    "Small",
    "Medium",
    "Large",
    "XLarge",
    "2XLarge",
    "3XLarge",
]

PipelineStatusType = Literal[
    "DRAFT",
    "RELEASED",
]


ProjectDataSortParameterType = Literal[
    "timeCreated",
    "-timeCreated",
    "timeModified",
    "-timeModified",
    "name",
    "-name",
    "path",
    "-path",
    "fileSizeInBytes",
    "-fileSizeInBytes",
    "status",
    "-status",
    "format",
    "-format",
    "dataType",
    "-dataType",
    "willBeArchivedAt",
    "-willBeArchivedAt",
    "willBeDeletedAt",
    "-willBeDeletedAt",
]


ProjectDataStatusValuesType = Literal[
    "PARTIAL",
    "AVAILABLE",
    "ARCHIVING",
    "ARCHIVED",
    "UNARCHIVING",
    "DELETING",
]


ProjectAnalysisSortParametersType = Literal[
    "reference",
    "-reference",
    "userReference",
    "-userReference",
    "pipeline",
    "-pipeline",
    "status",
    "-status",
    "startDate",
    "-startDate",
    "endDate",
    "-endDate",
    "summary",
    "-summary",
]


ProjectAnalysisStatusType = Literal[
    # In process
    "REQUESTED",
    "QUEUED",
    "INITIALIZING",
    "PREPARING_INPUTS",
    "IN_PROGRESS",
    "GENERATING_OUTPUTS",
    # Terminal Successful
    "SUCCEEDED",
    # Terminal Failed
    "ABORTING",
    "FAILED",
    "FAILED_FINAL",
    "ABORTED",
    # Unknown (maybe a nf thing)
    "AWAITING_INPUT",
]


ProjectAnalysisStepStatusType = Literal[
    "FAILED",
    "DONE",
    "RUNNING",
    "INTERRUPTED",
    "ABORTED",
    "WAITING",
]


WorkflowLanguageType = Literal[
    "CWL",
    "NEXTFLOW"
]


AnalysisLogStreamNameType = Literal[
    "stdout",
    "stderr",
]

JobStatusType = Literal[
    "INITIALIZED",
    "WAITING_FOR_RESOURCES",
    "RUNNING",
    "STOPPED",
    "SUCCEEDED",
    "PARTIALLY_SUCCEEDED",
    "FAILED",
]


UriType = Literal[
    "s3",
    "icav2",
]


DataTagType = Literal[
    "technical_tag",
    "user_tag",
    "connector_tag",
    "run_in_tag",
    "run_out_tag",
    "reference_tag",
]


AnalysisTagType = Literal[
    "user_tag",
    "technical_tag",
    "reference_tag",
]

ResourceType = Literal[
    "f1",
    "f2",
    "gpu",
    "software_only",
]


## Typed dicts
class AnalysisStepDict(TypedDict):
    name: str
    status: ProjectAnalysisStepStatusType
    queue_date: Optional[datetime]
    start_date: Optional[datetime]
    end_date: Optional[datetime]


## Unions


#!/usr/bin/env python
from __future__ import annotations

# Standard imports
import json
from datetime import datetime
from io import TextIOWrapper
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List, Union, Dict, Any, Optional, cast
import re

# Libica apis
from libica.openapi.v3.api.project_analysis_api import ProjectAnalysisApi
from libica.openapi.v3.api_client import ApiClient
from libica.openapi.v3.exceptions import ApiException
from pydantic import UUID4

# Wrapica imports
from ...literals import (
    ProjectAnalysisStatusType,
    ProjectAnalysisSortParametersType,
    AnalysisTagType,
    AnalysisStepDict,
    ProjectAnalysisStepStatusType,
    AnalysisLogStreamNameType,
)

# Libica models
from libica.openapi.v3.models import (
    AnalysisQueryParameters,
    AnalysisV3,
    AnalysisV4,
    AnalysisInput,
    AnalysisOutput,
    AnalysisOutputList,
    AnalysisStep,
    AnalysisStepLogs,
    CwlAnalysisInputJson,
    CwlAnalysisOutputJson,
)

# Local imports
from ...utils.globals import LIBICAV2_DEFAULT_PAGE_SIZE, IS_REGEX_MATCH
from ...utils.configuration import get_icav2_configuration
from ...utils.miscell import is_uuid_format, coerce_to_uuid4_obj
from ...utils.websocket_helpers import write_websocket_to_file, convert_html_to_text
from ...utils.logger import get_logger

AnalysisType = Union[AnalysisV3, AnalysisV4]

logger = get_logger()


def get_project_analysis_inputs(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str]
) -> List[AnalysisInput]:
    """
    Get the analysis inputs for a given analysis

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query

    :return: List of analysis inputs
    :rtype: List[`AnalysisInput <https://umccr.github.io/libica/openapi/v3/docs/AnalysisInput>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_project_analysis_inputs

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        workflow_inputs = get_project_analysis_inputs(project_id, analysis_id)

    """
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectAnalysisApi(api_client)

        # example passing only required values which don't have defaults set
        try:
            # Retrieve the outputs of an analysis.
            analysis_input_list: List[AnalysisInput] = api_instance.get_analysis_inputs(
                project_id=str(project_id),
                analysis_id=str(analysis_id)
            ).items
        except ApiException as e:
            logger.error("Exception when calling ProjectAnalysisApi->get_analysis_outputs: %s\n" % e)
            raise ApiException

    return analysis_input_list


def get_analysis_input_object_from_analysis_input_code(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str],
    analysis_input_code: str
) -> AnalysisInput:
    """
    Given an analysis input code for an analysis id, collect the analysis input object
    Confirm the input has either analysis or external data attributes

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query
    :param analysis_input_code: The analysis input code to query

    :return: The analysis input object
    :rtype: `AnalysisInput <https://umccr.github.io/libica/openapi/v3/docs/AnalysisInput>`_

    :raises: StopIteration, ValueError, ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_analysis_input_object_from_analysis_code

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"
        analysis_code = "run_folder"

        run_folder_input_data_id = get_analysis_input_object_from_analysis_code(
            project_id, analysis_id, analysis_code
        ).analysis_data[0].data_id
    """
    # Get analysis inputs
    analysis_input_list: List[AnalysisInput] = get_project_analysis_inputs(
        project_id,
        analysis_id
    )

    # Iterate through inputs to find the one we want
    try:
        input_obj: AnalysisInput = next(
            filter(
                lambda analysis_input_iter: analysis_input_iter.code == analysis_input_code,
                analysis_input_list
            )
        )
    except StopIteration:
        logger.error(f"Could not get {analysis_input_code} from analysis {analysis_id}")
        raise StopIteration

    if len(input_obj.analysis_data) == 0 and len(input_obj.external_data) == 0:
        logger.error(f"Expected analysis data or external data to be 1 but got {len(input_obj.analysis_data)}")
        raise ValueError

    return input_obj


def get_outputs_object_from_analysis_id(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str]
) -> List[AnalysisOutput]:
    """
    Query the outputs object from the analysis id

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query

    :return: List of analysis outputs
    :rtype: List[`AnalysisOutput <https://umccr.github.io/libica/openapi/v3/docs/AnalysisOutput>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python
    
        :linenos:
        from wrapica.project_analysis import get_outputs_object_from_analysis_id

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        workflow_outputs = get_outputs_object_from_analysis_id(project_id, analysis_id)
    """
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectAnalysisApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the outputs of an analysis
        api_response: AnalysisOutputList = api_instance.get_analysis_outputs(
            project_id=str(project_id),
            analysis_id=str(analysis_id)
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->get_analysis_outputs: %s\n" % e)
        raise ApiException

    return api_response.items


def get_analysis_output_object_from_analysis_output_code(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str],
    analysis_output_code: str
) -> AnalysisOutput:
    """
    Given an analysis code for an analysis id, collect the analysis output object

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query
    :param analysis_output_code: The analysis output code to collect

    :return: The analysis output object
    :rtype: `AnalysisOutput <https://umccr.github.io/libica/openapi/v3/docs/AnalysisOutput>`_

    :raises: StopIteration, ValueError, ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_analysis_output_object_from_analysis_code

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"
        analysis_code = "Output"

        analysis_folder_output_id = get_analysis_output_object_from_analysis_code(
            project_id, analysis_id, analysis_code
        ).data[0].data_id
    """
    analysis_output: List[AnalysisOutput] = get_outputs_object_from_analysis_id(
        project_id,
        analysis_id
    )
    try:
        output_obj: AnalysisOutput = next(
            filter(
                lambda analysis_iter: analysis_iter.code == analysis_output_code,
                analysis_output
            )
        )
    except StopIteration:
        logger.error(f"Could not get output item from analysis {analysis_id}")
        raise StopIteration

    if len(output_obj.data) == 0:
        logger.error(f"Expected analysis output data to be at least 1 but got {len(output_obj.data)}")
        raise ValueError

    return output_obj


def get_cwl_outputs_json_from_analysis_id(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str]
) -> Dict[str, Any]:
    """
    Query the outputs object from the analysis id

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query

    :return: List of analysis outputs
    :rtype: Dict[str, Any]

    :raises: ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_cwl_outputs_json_from_analysis_id

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        cwl_json_output = get_cwl_outputs_json_from_analysis_id(project_id, analysis_id)
    """
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectAnalysisApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the outputs of an analysis
        api_response: CwlAnalysisOutputJson = api_instance.get_cwl_output_json(
            project_id=str(project_id),
            analysis_id=coerce_to_uuid4_obj(analysis_id)
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->get_analysis_outputs: %s\n" % e)
        raise ApiException

    return json.loads(api_response.output_json)


def get_analysis_obj_from_analysis_id(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str]
) -> AnalysisV4:
    """
    Get an analysis object given a project id and analysis id

    :param project_id: The project context the analysis was run in
    :param analysis_id: The analysis id to query

    :return: The analysis object
    :rtype: `Analysis <https://umccr.github.io/libica/openapi/v3/docs/Analysis>`_

    :raises: ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_analysis_obj_from_analysis_id

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        analysis = get_analysis_obj_from_analysis_id(project_id, analysis_id)
    """
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Set as V4 headers
        api_client.set_default_header(
            header_name="Content-Type",
            header_value="application/vnd.illumina.v4+json"
        )
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v4+json"
        )
        # Create an instance of the API class
        api_instance = ProjectAnalysisApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve an analysis.
        api_response: AnalysisV4 = api_instance.get_analysis(
            project_id=str(project_id),
            analysis_id=str(analysis_id)
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->get_analysis: %s\n" % e)
        raise ApiException

    return api_response


def get_analysis_steps(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str],
    include_technical_steps: bool = False
) -> List[AnalysisStep]:
    """
    Get the workflow steps for a given analysis

    :param project_id:
    :param analysis_id:
    :param include_technical_steps:

    :return: List of analysis steps
    :rtype: List[`AnalysisStep <https://umccr.github.io/libica/openapi/v3/docs/AnalysisStep>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_analysis_steps

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        analysis_step_list = get_analysis_steps(project_id, analysis_id)
    """
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectAnalysisApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve the individual steps of an analysis.
        api_response = api_instance.get_analysis_steps(
            project_id=str(project_id),
            analysis_id=str(analysis_id)
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->get_analysis_steps: %s\n" % e)
        raise ApiException

    # Collect all steps
    analysis_steps: List[AnalysisStep] = api_response.items

    if not include_technical_steps:
        analysis_steps = list(
            filter(
                lambda step_iter: step_iter.technical is False,
                analysis_steps
            )
        )

    return analysis_steps


def get_analysis_log_from_analysis_step(
    analysis_step: AnalysisStep
) -> AnalysisStepLogs:
    """
    Get the logs for a given analysis step

    :param analysis_step:

    :return: Get the logs attribute of an analysis step
    :rtype: `AnalysisStepLogs <https://umccr.github.io/libica/openapi/v3/docs/AnalysisStepLogs>`_

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import get_analysis_steps

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        analysis_step_list = get_analysis_steps(project_id, analysis_id)

        step_logs = list(map(get_analysis_log_from_analysis_step, analysis_step_list))
    """
    return analysis_step.logs


def write_analysis_step_logs(
    project_id: Union[UUID4, str],
    step_logs: AnalysisStepLogs,
    log_name: AnalysisLogStreamNameType,
    output_path: Union[Path | TextIOWrapper],
    is_cwltool_log: Optional[bool] = False
) -> None:
    """
    Write the analysis step logs to a file

    :param project_id: The project id the analysis was run in
    :param step_logs: Required, the step logs object
    :param log_name: Required, One of stdout or stderr
    :param output_path: Required, The output path to write the file to
    :param is_cwltool_log: Required, If the log is a cwltool log we convert from html to text format

    :raises: ApiException, NotADirectoryError

    :return: None
    :rtype: None

    :Examples:

    .. code-block:: python

        :linenos:
        from pathlib import Path
        from wrapica.project_analysis import get_analysis_steps

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        # Get analysis steps and logs
        analysis_step_list = get_analysis_steps(project_id, analysis_id)
        step_logs = list(map(get_analysis_log_from_analysis_step, analysis_step_list))

        # Write first step logs to file
        write_analysis_step_logs(
            project_id=project_id,
            step_logs=step_logs[0],
            log_name="stderr",
            output_path=Path("stderr.log")
        )

    """
    # Local imports
    from ...project_data import read_icav2_file_contents

    # Check if we're getting our log from a stream
    is_stream = False
    log_stream = None
    log_data_id = ""

    # Initialise list of non empty log attributes
    non_empty_log_attrs = []

    # Check attributes of log obj
    for attr in dir(step_logs):
        if attr.startswith('_'):
            continue
        if getattr(step_logs, attr) is None:
            continue
        non_empty_log_attrs.append(attr)

    if log_name == "stderr":
        if hasattr(step_logs, "std_out_stream") and step_logs.std_out_stream is not None:
            is_stream = True
            log_stream = step_logs.std_out_stream
        elif hasattr(step_logs, "std_out_data") and step_logs.std_out_data is not None:
            log_data_id: str = str(step_logs.std_out_data.id)
        else:
            logger.error("Could not get either file output or stream of logs")
            logger.error(f"The available attributes were {', '.join(non_empty_log_attrs)}")
            raise AttributeError
    else:
        if hasattr(step_logs, "std_err_stream") and step_logs.std_err_stream is not None:
            is_stream = True
            log_stream = step_logs.std_err_stream
        elif hasattr(step_logs, "std_err_data") and step_logs.std_err_data is not None:
            log_data_id: str = str(step_logs.std_err_data.id)
        else:
            logger.error("Could not get either file output or stream of logs")
            logger.error(f"The available attributes were {', '.join(non_empty_log_attrs)}")
            raise AttributeError
    if is_stream:
        if is_cwltool_log:
            temp_html_obj = NamedTemporaryFile()
            write_websocket_to_file(
                log_stream,
                output_file=Path(temp_html_obj.name)
            )
            convert_html_to_text(Path(temp_html_obj.name), output_path)
        else:
            write_websocket_to_file(
                log_stream,
                output_file=output_path
            )
    else:
        read_icav2_file_contents(project_id, log_data_id, output_path)


def abort_analysis(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str],
) -> None:
    """
    Abort an analysis

    :param project_id: The project id the analysis was run in
    :param analysis_id: The analysis id to abort

    :raises: ApiException

    :return: None
    :rtype: None

    :Examples:

    .. code-block:: python

        :linenos:
        from pathlib import Path
        from wrapica.project_analysis import abort_analysis

        # Set params
        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id
        project_id = "project_id"
        analysis_id = "analysis_id"
    """
    # Get the configuration
    configuration = get_icav2_configuration()

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Force default headers for endpoints with a ':' in the name
        api_client.set_default_header(
            header_name="Content-Type",
            header_value="application/vnd.illumina.v3+json"
        )
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v3+json"
        )
        # Create an instance of the API class
        api_instance = ProjectAnalysisApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Abort an analysis.
        api_instance.abort_analysis(
            project_id=str(project_id),
            analysis_id=str(analysis_id)
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->abort_analysis: %s\n" % e)
        raise ApiException


def list_analyses(
    project_id: Union[UUID4, str],
    pipeline_id: Optional[Union[UUID4, str]] = None,
    user_reference: Optional[str] = None,
    status: Optional[Union[ProjectAnalysisStatusType, List[ProjectAnalysisStatusType]]] = None,
    creation_date_before: Optional[datetime] = None,
    creation_date_after: Optional[datetime] = None,
    modification_date_before: Optional[datetime] = None,
    modification_date_after: Optional[datetime] = None,
    sort: Optional[Union[ProjectAnalysisSortParametersType, List[ProjectAnalysisSortParametersType]]] = None,
    max_items: Optional[int] = None
) -> List[AnalysisV4]:
    """
    List analyses

    :param project_id: The project id
    :param pipeline_id: The pipeline id
    :param user_reference: The full user reference (regex optional)
    :param status: The status of the analysis
    :param creation_date_before: Return only analyses created before this date
    :param creation_date_after: Return only analyses created after this date
    :param modification_date_before: Return only analyses modified before this date
    :param modification_date_after: Return only analyses modified after this date
    :param sort: A parameter, or list of parameters to sort by
    :param max_items:

    :raises: ApiException

    :return: List of analyses
    :rtype: `List[Analysis] <https://umccr.github.io/libica/openapi/v3/docs/AnalysisV4/>`_

    :Examples:

    .. code-block:: python

        :linenos:
        from pathlib import Path
        from wrapica.project_analysis import list_analyses

        # Set params
        # Use wrapica.project.get_project_id_from_project_name
        # If you need to convert a project_name to a project_id
        project_id = "project_id"

        analysis_list = list_analyses(project_id)
    """
    # Get the configuration
    configuration = get_icav2_configuration()

    # Check parameters
    if user_reference is not None and IS_REGEX_MATCH.search(user_reference) is not None:
        user_reference_regex = re.compile(user_reference)
        user_reference = None
    else:
        user_reference_regex = None

    if status is not None and not isinstance(status, list):
        status = [status]

    if sort is not None:
        if not isinstance(sort, list):
            sort = [sort]

        # Complete a comma join of the sort parameters
        sort = ",".join(sort)

    # AnalysisQueryParameters
    analysis_query_parameters = {
        "status": status if status is not None else None,
        "user_reference": user_reference,
    }
    analysis_query_parameters = AnalysisQueryParameters(
        # Filter out None values and parse into query parameters
        **dict(filter(
            lambda kv_iter_: kv_iter_[1] is not None,
            analysis_query_parameters.items()
        ))
    )

    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = ProjectAnalysisApi(api_client)

    # Set page parameters
    if max_items is None:
        max_items = 0
    if not max_items == 0 and max_items < LIBICAV2_DEFAULT_PAGE_SIZE:
        page_size = max_items
    else:
        page_size = LIBICAV2_DEFAULT_PAGE_SIZE
    page_token = ""
    # We use page tokens if sort is None, otherwise we use page offsets
    if sort is not None:
        page_offset = 0
    else:
        page_offset = ""

    # Initialise list
    analysis_list: List[AnalysisV4] = []

    # Loop through the pages
    while True:
        # Attempt to collect all analyses
        try:
            api_response = api_instance.search_analyses(
                **dict(
                    filter(
                        lambda x: x[1] is not None,
                        {
                            "status": status,
                            "project_id": str(project_id),
                            "page_size": str(page_size),
                            "page_offset": str(page_offset),
                            "page_token": page_token,
                            "analysis_query_parameters": analysis_query_parameters,
                            "sort": sort
                        }.items()
                    )
                )
            )
        except ApiException as e:
            raise ValueError("Exception when calling ProjectAnalysisApi->get_project_data_list: %s\n" % e)

        # Extend items list
        analysis_list.extend(api_response.items)

        # Determine page iteration method by if we have a 'sort' parameter
        if sort is not None:
            # Check page offset and page size against total item count
            if page_offset + page_size >= api_response.total_item_count:
                break
            if not max_items == 0 and len(analysis_list) >= max_items:
                break

            # Continuing iteration
            page_offset += page_size
            # Add page size to page offset
            # But check if we're approaching the max_items
            if not max_items == 0 and len(analysis_list) + page_size > max_items:
                page_size = max_items - len(analysis_list)
        else:
            # Check if there is a next page
            if api_response.next_page_token is None or api_response.next_page_token == "":
                break
            page_token = api_response.next_page_token

    # Before we return the list, we filter
    # pipeline_id
    # creation_date_after
    # creation_date_before
    # modification_date_after
    # modification_date_before
    analysis_list = list(
        filter(
            lambda analysis_iter: (
                (
                    pipeline_id is None or
                    str(analysis_iter.pipeline.id) == str(pipeline_id)
                ) and
                (
                    user_reference_regex is None or
                    user_reference_regex.match(analysis_iter.user_reference) is not None
                ) and
                (
                    creation_date_after is None or
                    analysis_iter.creation_date >= creation_date_after
                ) and
                (
                    creation_date_before is None or
                    analysis_iter.creation_date <= creation_date_before
                ) and
                (
                    modification_date_after is None or
                    analysis_iter.modification_date >= modification_date_after
                ) and
                (
                    modification_date_before is None or
                    analysis_iter.modification_date <= modification_date_before
                )
            ),
            analysis_list
        )
    )

    return analysis_list


def get_cwl_analysis_input_json(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str]
) -> Dict:
    """
    Get the CWL Analysis Input JSON

    :param project_id: The project id the analysis was run in
    :param analysis_id:  The analysis id

    :return: The CWL Analysis Input JSON As a dictionary
    :rtype: Dict

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        import json
        from wrapica.project_analysis import get_cwl_analysis_input_json

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        # Print the input json
        print(
            json.dumps(
                get_cwl_analysis_input_json(project_id, analysis_id),
                indent=4
            )
        )

    """
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectAnalysisApi(api_client)

    try:
        # Retrieve the input json of a CWL analysis.
        api_response: CwlAnalysisInputJson = api_instance.get_cwl_input_json(
            project_id=project_id,
            analysis_id=analysis_id
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->get_cwl_input_json: %s\n" % e)
        raise ApiException

    return json.loads(api_response.input_json)


def get_cwl_analysis_output_json(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str]
) -> Dict:
    """
    Get the CWL Analysis Input JSON

    :param project_id: The project id the analysis was run in
    :param analysis_id: The analysis id

    :return: The CWL Analysis Output JSON As a dictionary
    :rtype: Dict

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        import json
        from wrapica.project_analysis import get_cwl_analysis_output_json

        # Set params
        project_id = "project_id"
        analysis_id = "analysis_id"

        # Print the output json
        print(
            json.dumps(
                get_cwl_analysis_output_json(project_id, analysis_id),
                indent=4
            )
        )

    """
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectAnalysisApi(api_client)

    try:
        # Retrieve the input json of a CWL analysis.
        api_response: CwlAnalysisOutputJson = api_instance.get_cwl_output_json(
            str(project_id),
            coerce_to_uuid4_obj(analysis_id)
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->get_cwl_output_json: %s\n" % e)
        raise ApiException

    return json.loads(api_response.output_json)


def analysis_step_to_dict(analysis_step: AnalysisStep) -> AnalysisStepDict:
    """
    Convert an analysis step object to a dictionary

    Returns the following attributes
    * name: str -> remove preceding '#' values
    * status: WorkflowStep.Status -> The workflow step status
    * queue_date: Datetime object
    * start_date: Datetime object
    * end_date: Datetime object

    :param analysis_step: The AnalysisStep object
    :return: The analysis step as a dict
    :rtype: Dict
    """
    return {
        "name": analysis_step.name.split("#", 1)[-1],
        "status": cast(ProjectAnalysisStepStatusType, analysis_step.status),
        "queue_date": analysis_step.queue_date if hasattr(analysis_step, "queue_date") else None,
        "start_date": analysis_step.start_date if hasattr(analysis_step, "start_date") else None,
        "end_date": analysis_step.end_date if hasattr(analysis_step, "end_date") else None
    }


def get_analysis_obj_from_user_reference(
    project_id: Union[UUID4, str],
    user_reference: str
) -> AnalysisType:
    """
    Given a user reference, get the analysis object

    Will fail if more than one analysis is found for a given user reference.
    Will also fail if no analysis is found for the user reference.

    :param project_id:
    :param user_reference:
    :return:
    """

    # List analysis filtering on the user reference
    # List analysis filtering on the user reference
    analysis_list = list_analyses(
        project_id=project_id,
        user_reference=user_reference
    )

    if len(analysis_list) == 0:
        logger.error("Could not find analysis id from user reference")
        raise ValueError

    if not len(analysis_list) == 1:
        logger.error(
            f"Got {len(analysis_list)} analyses from user reference, "
            f"cannot coerce user reference to analysis id"
        )
        raise ValueError

    return analysis_list[0]


def coerce_analysis_id_or_user_reference_to_analysis_obj(
    project_id: Union[UUID4, str],
    analysis_id_or_user_reference: Union[Union[UUID4, str], str]
) -> AnalysisType:
    """
    Given either an analysis id or user reference, coerce to an analysis object

    :param project_id:
    :param analysis_id_or_user_reference:
    :return: The analysis object
    :rtype: `Analysis <https://umccr.github.io/libica/openapi/v3/docs/Analysis>`_

    :raises: ValueError

    :Examples:

    .. code-block:: python

        :linenos:
        from wrapica.project_analysis import coerce_analysis_id_or_user_reference_to_analysis_obj

        # Set params
        analysis_id_or_user_reference = "analysis_id"

        analysis = coerce_analysis_id_or_user_reference_to_analysis_obj(analysis_id_or_user_reference)
    """

    # For an analysis id
    if is_uuid_format(analysis_id_or_user_reference):
        return get_analysis_obj_from_analysis_id(
            project_id=project_id,
            analysis_id=analysis_id_or_user_reference
        )

    return get_analysis_obj_from_user_reference(
        project_id=project_id,
        user_reference=analysis_id_or_user_reference
    )


def coerce_analysis_id_or_user_reference_to_analysis_id(
    project_id: Union[UUID4, str],
    analysis_id_or_user_reference: str
) -> str:
    """
    Given either an analysis id or user reference, coerce to an analysis id

    :param project_id:
    :param analysis_id_or_user_reference:
    :return:
    """

    if is_uuid_format(analysis_id_or_user_reference):
        return analysis_id_or_user_reference

    return str(
        get_analysis_obj_from_user_reference(
            project_id=project_id,
            user_reference=analysis_id_or_user_reference
        ).id
    )


def update_analysis_obj(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str],
    analysis_obj: AnalysisType
) -> AnalysisType:
    """
    Given an analysis id, update the analysis object
    :param project_id:
    :param analysis_id:
    :param analysis_obj:
    :return:
    """
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectAnalysisApi(api_client)

    try:
        # Retrieve the input json of a CWL analysis.
        api_response: AnalysisType = api_instance.update_analysis(
            project_id=str(project_id),
            analysis_id=str(analysis_id),
            analysis_v4=(
                analysis_obj
                if isinstance(analysis_obj, AnalysisV4)
                else get_analysis_obj_from_analysis_id(
                    project_id=project_id,
                    analysis_id=analysis_id
                )
            )
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectAnalysisApi->update_analysis: %s\n" % e)
        raise e

    return api_response


def add_tag_to_analysis(
    project_id: Union[UUID4, str],
    analysis_id: Union[UUID4, str],
    tag: str,
    tag_type: AnalysisTagType
):
    # Check project id and analysis id are in uuid formats
    if not is_uuid_format(project_id):
        raise ValueError("Project id is not in UUID format")

    if not is_uuid_format(analysis_id):
        raise ValueError("Analysis id is not in UUID format")

    # Get the current analysis object
    analysis_obj: AnalysisType = get_analysis_obj_from_analysis_id(
        project_id=project_id,
        analysis_id=analysis_id
    )

    if tag_type == "user_tag":
        analysis_obj.tags.user_tags.append(tag)
    elif tag_type == "technical_tag":
        analysis_obj.tags.technical_tags.append(tag)
    elif tag_type == "reference_tag":
        analysis_obj.tags.reference_tags.append(tag)
    else:
        raise ValueError("Tag type not recognised")

    # Update the analysis object
    analysis_obj = update_analysis_obj(
        project_id=project_id,
        analysis_id=analysis_id,
        analysis_obj=analysis_obj
    )

    return analysis_obj





# Standard imports
from typing import Union

# Libica model imports
from libica.openapi.v3.models import (
    PipelineFile,
    PipelineV3,
    PipelineV4
)

# Local imports
from .functions.pipelines_functions import (
    get_pipeline_obj_from_pipeline_id,
    get_pipeline_obj_from_pipeline_code,
    coerce_pipeline_id_or_code_to_pipeline_obj,
    coerce_pipeline_id_or_code_to_pipeline_id,
    list_all_pipelines,
    download_pipeline_file,
    list_pipeline_files,
    download_pipeline_to_directory,
    download_pipeline_to_zip,
    get_cwl_obj_from_pipeline_id
)

PipelineType = Union[PipelineV3, PipelineV4]

__all__ = [
    # Libica imports
    'PipelineType',
    'PipelineV3',
    'PipelineV4',
    'PipelineFile',
    # Local functions
    'get_pipeline_obj_from_pipeline_id',
    'get_pipeline_obj_from_pipeline_code',
    'coerce_pipeline_id_or_code_to_pipeline_obj',
    'coerce_pipeline_id_or_code_to_pipeline_id',
    'list_all_pipelines',
    'download_pipeline_file',
    'list_pipeline_files',
    'download_pipeline_to_directory',
    'download_pipeline_to_zip',
    'get_cwl_obj_from_pipeline_id',
]
